from numpy import sqrt
from numpy import pi
import numpy as np
from .stonefence import ORDER_L


def a(ell, m):
    if ((3*(ell-m)*(ell+m))/(4*pi*(2*ell-1)*(2*ell+1))<0):
        print(ell, m)
        raise ValueError("negative value")

    return np.sqrt((3*(ell-m)*(ell+m))/(4*pi*(2*ell-1)*(2*ell+1)))

def apm(ell, m, pm):
    if 3*(ell+m)*(ell+m+pm)/(8*pi*(2*ell-1)*(2*ell+1)) < 0:
        print(ell, m, pm)
        raise ValueError("negative value")
    return np.sqrt(3*(ell+m)*(ell+m+pm)/(8*pi*(2*ell-1)*(2*ell+1)))
def ap(ell, m):
    return apm(ell, m, 1)
def am(ell, m):
    return apm(ell, m, -1)

Y00 = 1.0/sqrt(4*pi)





def YYYreal(m1,L,M,m2, ORDER_L):
    '''
    YYYreal_{l_1,l_2} = \\int d\\vec{u} Y_{l_1, m1} Y_{L,M} Y_{l2,m2}

    Not able to handle all combiniations of inputs.   Can only handle
    Y_{l_1,0} Y_{2,M} Y_{l_2,M}
    or
    Y_{l_1,M} Y_{2,M} Y_{l_2,0}
    or
    Y_{l_1,m_1} Y_{1,M} Y{l_2,m_2}
    or
    Y_{l_1,m_1} Y_{0,0} Y{l_2,m_2}

    '''
    
    assert type(ORDER_L) == type(20), "ORDER_L must be an integer"
    assert abs(M) <= L, "l must be >= |m|"

    if m1 == 0 and L==2 and M==m2:
        return YYY_0mm(M, ORDER_L)

    if L==1:
        return YYY_l1(m1, M, m2, ORDER_L)

    if m2 ==0 and L==2 and M==m1:
        return np.transpose( YYY_0mm(M, ORDER_L) )

    if L==0 and M==0:
        if m1 == m2:
            return np.identity(ORDER_L)/np.sqrt(np.pi*4)
        else:
            return np.zeros((ORDER_L,ORDER_L))
    
    print(m1,L,M,m2)
    raise ValueError("Case not handeled")


# -----  Store precalculate Vaues ------
saveYYY = {}
def YYY(m1, L, M, m2, ORDER_L=ORDER_L):
    """
    Same as YYYreal
    """
    key = (m1, L,M, m2)
    if key in saveYYY:
        return saveYYY.get(key)
    else:
        value = YYYreal(m1, L, M, m2, ORDER_L)
        saveYYY[key] = value
        return value

def YYY_0mm(m, ORDER_L):
    """
    Y_{ell1,0} Y_{2,m} Y_{ell2,m}
    """
    out = np.zeros((ORDER_L, ORDER_L))
    if m==0:
        for l1 in range(0, ORDER_L-2):
            out[l1, l1+2] = a(l1+1,0)*a(l1+2,0)/a(2,0)
        for l1 in range(0, ORDER_L):
            out[l1,l1] = (a(l1+1,0)**2 - a(1,0)*Y00 + a(l1,0)**2)/a(2,0)
        for l1 in range(2, ORDER_L):
            out[l1, l1-2] = a(l1,0)*a(l1-1,0)/a(2,0)

    elif abs(m) == 1:
        for l1 in range(0, ORDER_L-2):
            out[l1, l1+2] = ap(l1+1,0)*a(l1+2,1)/a(2,1) 
        for l1 in range(abs(m), ORDER_L):
            out[l1, l1] = (ap(l1+1,0)*a(l1+1,1)-am(l1,0)*a(l1,1))/a(2,1)
        for l1 in range(abs(m)+2, ORDER_L):
            out[l1, l1-2] = -am(l1,0)*a(l1-1,1)/a(2,1)
    elif abs(m) == 2:
        for l1 in range(0, ORDER_L-2):
            out[l1, l1+2] = ap(l1+1,0)*ap(l1+2,1)/ap(2,1)
        for l1 in range(abs(m), ORDER_L):
            out[l1, l1] = -(ap(l1+1,0)*am(l1+1,-1)+am(l1,0)*ap(l1,1))/ap(2,1)
        for l1 in range(abs(m)+2, ORDER_L):
            out[l1, l1-2] = am(l1,0)*am(l1-1,-1)/ap(2,1)
    else:
        raise ValueError("m out or range")
    return out
    

def f(m):
    if m<0:
        return -1.0/sqrt(2)
    if m==0:
        return 0.0
    if m==1:
        return 1.0
    if m>1:
        return 1.0/sqrt(2)

def fzz(m):
    if m<0:
        return -1.0/sqrt(2)
    if m==0:
        return 0.0
    if m==1:
        return 0.0
    if m>1:
        return 1.0/sqrt(2)

def ftt(m):
    if m<-1:
        return -1/sqrt(2)
    if m==-1:
        return -1.0
    if m==0:
        return 1.0
    if m>0:
        return 1.0/sqrt(2)

def YYY_l1(m1,m,m2,ORDER_L):
    """
    Y_{l1,m1} Y_{1,m} Y{l2,m2}
    """
    out = np.zeros((ORDER_L, ORDER_L))
    if m == 0:
        if m1 != m2:
            return out
        for l1 in range(0,ORDER_L-1):
            if l1<abs(m1) or l1+1<abs(m2):
                continue
            out[l1,l1+1] = a(l1+1,m1)
        for l1 in range(1,ORDER_L):
            if l1<abs(m1) or l1-1<abs(m2):
                continue
            out[l1,l1-1] = a(l1,m1)
    if m == 1:
        if m2 == m1+1:
            for l1 in range(0,ORDER_L-1):
                if l1<abs(m1) or l1+1<abs(m2):
                    continue
                out[l1,l1+1] = ap(l1+1,m1)*f(m1+1)
            for l1 in range(1,ORDER_L):
                if l1<abs(m1) or l1-1<abs(m2):
                    continue
                out[l1,l1-1] = -ap(l1,-m1-1)*f(m1+1)
        if m2 == m1-1:
            for l1 in range(0,ORDER_L-1):
                if l1<abs(m1) or l1+1<abs(m2):
                    continue
                out[l1,l1+1] = -ap(l1+1,-m1)*f(m1)
            for l1 in range(1,ORDER_L):
                if l1<abs(m1) or l1-1<abs(m2):
                    continue
                out[l1,l1-1] = ap(l1,m1-1)*f(m1)
    if m == -1:
        if m2 == -m1+1:
            for l1 in range(0,ORDER_L-1):
                if l1<abs(m1) or l1+1<abs(m2):
                    continue
                out[l1,l1+1] = ap(l1+1,-m1)*fzz(m1)
            for l1 in range(1,ORDER_L):
                if l1<abs(m1) or l1-1<abs(m2):
                    continue
                out[l1,l1-1] = -ap(l1,m1-1)*fzz(m1)
        if m2 == -m1-1:
            for l1 in range(0,ORDER_L-1):
                if l1<abs(m1) or l1+1<abs(m2):
                    continue
                out[l1,l1+1] = ap(l1+1,m1)*ftt(m1)
            for l1 in range(1,ORDER_L):
                if l1<abs(m1) or l1-1<abs(m2):
                    continue
                out[l1,l1-1] = -ap(l1,-m1-1)*ftt(m1)
    return out


#raise Exception('exit')
# ------------------  Testing ----------------------------------

from  scipy.special import sph_harm as Y
from scipy.integrate import dblquad as intt

def ind_sphere(fun):
    """Integrate function on sphere.

    Args:
        fun (callable) : fun(theta, phi)
    """
    def refun(phi, theta):
        return fun(theta, phi)*np.sin(theta)
    def gfun(x):
        return 0.0
    def hfun(x):
        return 2*np.pi
    return intt(refun, 0, np.pi, gfun, hfun)[0]

def YR(m, ell, phi, theta):
    '''
    Real spherical harmonic function.
    '''
    if m<0:
        out = (1.0j/sqrt(2))*(Y(m, ell, phi, theta)
                               - (-1)**m * Y(-m, ell, phi, theta))
    if m==0:
        out = Y(m, ell, phi, theta)
    if m>0:
        out = (1.0/sqrt(2))*(Y(-m, ell, phi, theta) 
                             + (-1)**m * Y(m, ell, phi, theta))

    if np.isnan(np.real(out)):
        import pdb
        pdb.set_trace()

    return np.real(out)

def intYYYnum(l1, m1, L, M, l2, m2):
    '''
    Numerical integral over 3 spherical harmics
    '''
    def num_YYY(theta, phi):
        return YR(m1, l1, phi, theta)*\
               YR(M, L, phi, theta)*\
               YR(m2, l2, phi, theta)
    return ind_sphere(num_YYY)



def testYYYreal():
    ORDER_L=20
    values = {}
    L = 1
    for M in [-1,0,1]:
        elmax=5
        mtrx = YYYreal(0, L, M, M, ORDER_L)
        for l1 in range(0,elmax):
            for l2 in range(abs(M),elmax):
                values[(l1,0,L,M,l2,M)] = mtrx[l1,l2]
                # compair mtrx[l1,l2] with Y_{l1,0} Y_{L,M} Y_{l2,M}
    
    L = 2
    for M in [-2,-1,0,1,2]:
        m1=0
        m2=M
        mtrx = YYYreal(m1,L,M,m2,ORDER_L)
        for l1 in range(abs(m1),elmax):
            for l2 in range(abs(m2),elmax):
                values[(l1,m1,L,M,l2,m2)] = mtrx[l1,l2]
                #compair mtrx[l1,l2] with Y_{l1,m1} Y_{L,M} Y_{l2,m2}

    for key, value in values.items():
        print("-------------------")
        print(key)
        num = intYYYnum(*key)
        diff = abs(num - value)
        print(value)
        print(num)
        if diff > 0.00001 or np.isnan(diff):
            print("diff")
            print(diff)
            print("num")
            print(num)
            print("value")
            print(value)
            raise ValueError("Doesn't Match")
#testYYYreal()



import numpy as np
from numba import jit

@jit(nopython=True)
def cot_plane(u, unit=True, tol=1e-8):
    """Return an arbitrary basis to the cotangent plane given the tangent
    vector (assumed unit vector unless otherwise specified)."""
    if not unit:
        u = u/np.linalg.norm(u)
    n1 = np.array([0.0, 0.0, 1.0])
    n1 = n1 - (n1@u)*u
    mag = np.linalg.norm(n1)
    # if we accidentally chose vector parallel to u
    if np.abs(mag) < tol:
        n1 = np.array([0.0, 1.0, 0.0])
        n1 = n1 - (n1@u)*u
        mag = np.linalg.norm(n1)
    n1 = n1/mag
    # n2 = np.cross(n1, u)
    # for jit
    n2 = np.array([n1[1]*u[2] - n1[2]*u[1],
                   n1[2]*u[0] - n1[0]*u[2],
                   n1[0]*u[1] - n1[1]*u[0]])

    n2 = n2/np.linalg.norm(n2)
    return n1, n2

def Ry(theta):
    r"""Rotation matrix about the z axis with angle theta.

    Notes
    -----
    ..math::

        \frac{1}{\sqrt{3}}
        \begin{bmatrix}
            np.cos(theta) & 0 & -np.sin(theta) \\
                        0 & 1 &              0 \\
            np.sin(theta) & 0 &  np.cos(theta)
        \end{bmatrix}
    """
    return np.array([[np.cos(theta), 0, -np.sin(theta)],
                     [            0, 1,              0],
                     [np.sin(theta), 0,  np.cos(theta)]])


def Rz(theta):
    r"""Rotation matrix about the z axis with angle theta.

    Notes
    -----
    ..math::

        \frac{1}{\sqrt{3}}
        \begin{bmatrix}
            np.cos(theta) & -np.sin(theta) & 0 \\
            np.sin(theta) &  np.cos(theta) & 0 \\
                        0 &             0  & 1
        \end{bmatrix}
    """
    return np.array([[np.cos(theta), -np.sin(theta), 0],
                    [np.sin(theta),  np.cos(theta), 0],
                    [            0,             0,  1]])




import numpy as np
import matplotlib
from matplotlib import pyplot as plt
from multiprocessing import Pool
import pdb
import seaborn as sns
import pickle

from .stonefence import ORDER_L
from .rigidrod import rodFrank
from .frank import *

def multi_process_fun(inputs):
    "pass through for muiltiproccessing"
    return get_Frank_values(*inputs, gammaNmax=10**3)
    
def plot_Frank(log_gamma=True, MultiProcess=True, npts=24, N=1.0,
               gammaN=False, maxgamma=35, nthreads=25):
    """Calculate Frank Elastic data for range of gamma values.

    Args:
        log_gamma (bool): Space gamma logrithmeicly
        MultiProcess (bool): Use multiple threads
        npts (int): number gamma values
        N (real): Kuhn lenths in polymer
        gammaN (bool): Devide gamma by N.  I.e. nondimensionalized by polyer length
        nthreads (int): Number of threads to use
    """
    if log_gamma:
        gammas = np.logspace(0.0,3.5,npts)
    else:
        gammas = np.linspace(0,maxgamma,npts)
    if gammaN:
        gammas=gammas/N
   
    inputs = [(gamma, N) for gamma in gammas]
    if MultiProcess:
        #if __name__ == '__main__':
        with Pool(nthreads) as my_pool:
            results = my_pool.map(multi_process_fun, inputs)
    else:
        results = []
        for inputt in inputs:
            print("Done with %d of %d"%(ii, len(gammas)))
            results.append(get_Frank_values(*inputt))
    
    # Now we need to reorganize the output for plotting
    toplot={}
    for key in results[0].keys():
        toplot[key] = np.zeros(npts)
        for ii, gamma in enumerate(gammas):
            if key not in results[ii].keys():
                toplot[key][ii]=np.NaN
                continue
            if key not in toplot.keys():
                toplot[key]=np.zeros(npts)*np.NaN
            try:
                toplot[key][ii] = results[ii][key]
            except:
                pdb.set_trace()

    toplot["gammas"]=gammas
    return toplot

def get_data_for_Splot(name, nthreads=25, N=0.05, npts=45):
    """Get data for plot_S

    Args:
        name (string): Where to save result.
        nthreads (int): Number of computational threads
        N (real): Kuhn lenths in polymer
        npts (int): number gamma values
    """
    gammaN=False
    data = plot_Frank(N=N, log_gamma=True, MultiProcess=True,
                      npts=npts, gammaN=gammaN, nthreads=nthreads)
    data["npts"]=45
    data["gammaN"]=gammaN
    data["N"]=N
    pickle.dump(data,open(name,"wb"))


def plot_S(name):
    """Plot S over diffetn gamma values.
    Args:
        name (string): File name to load data from.
    """
    data = pickle.load(open(name,"rb"))
    for (l1, l2, m1, m2) in [(0,0,0,0), (2,2,0,0), (0,2,0,0), (0,2,0,0), (2,2,-1,-1),
                             (2,2,1,1), (2,2,-2,-2), (2,2,2,2)]:
        if m1<0:
            linetype = "o-"
        else:
            linetype = ".-"
        label="l1=%d, l2=%d, m1=%d, m2=%d"%(l1, l2, m1, m2)
        llmm = "S'"+str(l1)+str(l2)+str(m1)+str(m2)
        plt.plot(data["gammas"], data[llmm], linetype, label=label)
    plt.xscale("log")
    plt.xlim([1,1000])
    plt.ylim([0,1.05])
    plt.xlabel("gamma")
    plt.legend()
    plt.savefig("Splot.pdf")
    plt.show()

def get_Multi_frank_data(log_gamma=False, npts=48,
                     gammaN=False, maxgamma=35, saveas="Frank.p",
                     nthreads=25, Nset=None):
    """Calculate Frank elastic constants for a range of gamma and N.

    Args:
        log_gamma (bool): Space gamma logrithmeicly
        npts (int): number gamma values
        N (real): Kuhn lenths in polymer
        saveas= Where to save results.
        gammaN (bool): Devide gamma by N.  I.e. nondimensionalized by polyer length
        nthreads (int): Number of threads to use
        Nset (iterable): Where to save results
    """
    if Nset is None:
        Nset = [0.0316, 0.1, 0.316, 1.0, 3.16, 10.0, 31.6, 100.0]
    FofN={}

    for N in Nset:
        FofN[N] = plot_Frank(N=N, log_gamma=log_gamma,
                             MultiProcess=True, npts=npts, gammaN=gammaN,
                            maxgamma=maxgamma, nthreads=nthreads)
    FofN["ORDER_L"]=ORDER_L
    FofN["npts"]=npts
    FofN["log_gamma"]=log_gamma
    FofN["Nset"]=Nset
    FofN["maxgamma"]=maxgamma
    FofN["gammaN"]=gammaN
    pickle.dump(FofN, open(saveas,"wb"))

def Multi_frank_plot(load=None, logy=False):
    """Plot data from get_Multi_frank_data

    Args:
        load (string): File to load
        logy (bool): Log y axis
    """
    FofN=pickle.load(open(load,"rb"))
    Nset = FofN["Nset"]
    log_gamma=FofN['log_gamma']
    gammaN=FofN['gammaN']
    maxgamma=FofN['maxgamma']
    if load in {"./Frank_as_gammaN.p","./Frank_as_gamma.p"}:
        for N in Nset:
            if type(N) != type(0.1):
                continue
            FofN[N]["K'splay"]=FofN[N]["K'splay"]/2.0
            FofN[N]["K'twist"]=FofN[N]["K'twist"]/2.0
            FofN[N]["K'bend"]=FofN[N]["K'bend"]/2.0
    
    for N in Nset:
        if type(N) != type(0.1):
            continue
        a_values = FofN[N]["aLAf"]
        temp=[]
        for idx, val in enumerate(a_values):
            if np.isnan(val):
                continue
            temp.append((val,idx))
        val, idx = min(temp)

        FofN[N]["a_min"] = val
        FofN[N]["idx_a_min"] = idx
    colors = sns.color_palette("bright",len(Nset))

    if (gammaN):
        # ----------------------------------------------------------#
        #                       Low N
        # ----------------------------------------------------------#
        rod_data = rodFrank()

        idx = rod_data["idx_a_min"]
        plt.plot(rod_data["gammas*N"][idx:], rod_data["K_splay"][idx:], "--", color="blue")
        plt.plot(rod_data["gammas*N"][idx:], rod_data["K_bend"][idx:], "--", color="red")
        plt.plot(rod_data["gammas*N"][idx:], rod_data["K_twist"][idx:], "--", color="green")
        for N in Nset:
            if N==0.0316:
                alpha=1.0
                labs=["splay","bend","twist"]
            else:
                alpha=0.3
                labs=[None, None, None]
            if type(N) != type(0.1):
                continue
            gammas=FofN[N]["gammas"]
            idx = FofN[N]["idx_a_min"]
            plt.plot(gammas[idx:]*N, FofN[N]["K'splay"][idx:], color="blue", alpha=alpha,
                     label=labs[0])
            plt.plot(gammas[idx:]*N, FofN[N]["K'bend"][idx:], color="red", alpha=alpha,
                     label=labs[1])
            plt.plot(gammas[idx:]*N, FofN[N]["K'twist"][idx:], color="green", alpha=alpha,
                     label=labs[2])
        if log_gamma:
            plt.xscale("log")
        if logy:
            plt.yscale("log")
        plt.xlabel("gamma*N")
        plt.ylabel("K'")
        plt.title("Low N")
        plt.legend()
        plt.show()

        idx = rod_data["idx_a_min"]
        plt.plot(rod_data["aLAf"][idx:], rod_data["K_splay"][idx:], "--", color="blue")
        plt.plot(rod_data["aLAf"][idx:], rod_data["K_bend"][idx:], "--", color="red")
        plt.plot(rod_data["aLAf"][idx:], rod_data["K_twist"][idx:], "--", color="green")

        for N in Nset:
            #if N==0.0316:
            if N==1.0:
                alpha=0.5
                labs=["splay","bend","twist"]
            else:
                alpha=0.5
                labs=[None, None, None]
            if type(N) != type(0.1):
                continue
            a_values = FofN[N]["aLAf"]
            idx = FofN[N]["idx_a_min"]
            plt.plot(a_values[idx:], FofN[N]["K'splay"][idx:], color="blue", alpha=alpha,
                     label=labs[0])
            plt.plot(a_values[idx:], FofN[N]["K'bend"][idx:], color="red", alpha=alpha,
                     label=labs[1])
            plt.plot(a_values[idx:], FofN[N]["K'twist"][idx:], color="green", alpha=alpha,
                     label=labs[2])
        if log_gamma:
            plt.xscale("log")
        if logy:
            plt.yscale("log")
        plt.xlabel("a(LAf)")
        plt.ylabel("K'")
        plt.title("Low N")
        plt.xlim([5,30])
        plt.ylim([0,2.5])
        plt.legend()
        plt.savefig("Low_N_Kplot.pdf")
        plt.show()
        
        plt.plot(rod_data["aLAf"], rod_data["gammas*N"], "--", color="black",
                 label=["N=0"])
        for iN, N in enumerate(Nset):
            if type(N) != type(0.1):
                continue
            if N > 5:
                continue
            gammas=FofN[N]["gammas"]
            a_values = FofN[N]["aLAf"]
            plt.plot(a_values, gammas*N, label=str(N), color=colors[iN])
        if log_gamma:
            plt.xscale("log")
            plt.yscale("log")
        plt.ylabel("gamma*N")
        plt.xlabel("a(LAf)")
        plt.xlim([0,70])
        plt.ylim([0,35])
        plt.legend()
        plt.savefig("gamma_lowN.pdf")
        plt.show()
    else:
        # ----------------------------------------------------------#
        #                       High N
        # ----------------------------------------------------------#
        for N in Nset:
            #if N==100.:
            if N==1.0:
                alpha=1.0
                labs=["splay","bend","twist"]
            else:
                alpha=0.3
                labs=[None, None, None]
            if type(N) != type(0.1):
                continue
            gammas=FofN[N]["gammas"]
            plt.plot(gammas, FofN[N]["K'splay"]*N, color="blue", alpha=alpha,
                     label=labs[0])
            plt.plot(gammas, FofN[N]["K'bend"]*N, color="red", alpha=alpha,
                     label=labs[1])
            plt.plot(gammas, FofN[N]["K'twist"]*N, color="green", alpha=alpha,
                     label=labs[2])
        if log_gamma:
            plt.xscale("log")
        if logy:
            plt.yscale("log")
        plt.xlabel("gamma")
        plt.ylabel("K'*N")
        plt.title("High N")
        plt.xlim([0,100])
        plt.ylim([0,5])
        plt.legend()
        plt.show()

        for N in Nset:
            #if N==100.:
            if N==1.0:
                alpha=1.0
                labs=["splay","bend","twist"]
            else:
                alpha=0.3
                labs=[None, None, None]
            if type(N) != type(0.1):
                continue
            a_values = FofN[N]["a2lpAf"]
            idx = FofN[N]["idx_a_min"]
            plt.plot(a_values[idx:], FofN[N]["K'splay"][idx:]*N, color="blue", alpha=alpha,
                     label=labs[0])
            plt.plot(a_values[idx:], FofN[N]["K'bend"][idx:]*N, color="red", alpha=alpha,
                     label=labs[1])
            plt.plot(a_values[idx:], FofN[N]["K'twist"][idx:]*N, color="green", alpha=alpha,
                     label=labs[2])
        if log_gamma:
            plt.xscale("log")
        if logy:
            plt.yscale("log")
        plt.xlabel("a(2lpAf)")
        plt.ylabel("K'*N")
        plt.title("High N")
        plt.xlim([0,100])
        plt.ylim([0,5])
        plt.legend()
        plt.show()

        for iN, N in enumerate(Nset):
            if type(N) != type(0.1):
                continue
            if N<0.3:
                continue
            gammas=FofN[N]["gammas"]
            a_values = FofN[N]["a2lpAf"]
            plt.plot(a_values, gammas, label=str(N), color=colors[iN])
        if log_gamma:
            plt.xscale("log")
            plt.yscale("log")
        plt.legend()
        plt.ylabel("gamma")
        plt.xlabel("a(2lpAf)")
        plt.xlim([18,40])
        plt.ylim([0.0,30])
        plt.savefig("gamma_highN.pdf")
        plt.show()
        
def seperate_bend_twist_splay(load, gammaN=False, logy=False):
    """Plot data from get_Multi_frank_data seperatly

    Args:
        load (string): File to load
        logy (bool): Log y axis
    """
    font={'size':12}
    matplotlib.rc('font',**font)
    FofN=pickle.load(open(load,"rb"))
    Nset = FofN["Nset"]
    log_gamma = FofN["log_gamma"]
    if load in {"./Frank_as_gammaN.p","./Frank_as_gamma.p"}:
        for N in Nset:
            if type(N) != type(0.1):
                continue
            FofN[N]["K'splay"]=FofN[N]["K'splay"]/2.0
            FofN[N]["K'twist"]=FofN[N]["K'twist"]/2.0
            FofN[N]["K'bend"]=FofN[N]["K'bend"]/2.0
    for N in Nset:
        if type(N) != type(0.1):
            continue
        a_values = FofN[N]["aLAf"]
        temp=[]
        for idx, val in enumerate(a_values):
            if np.isnan(val):
                continue
            temp.append((val,idx))
        val, idx = min(temp)

        FofN[N]["a_min"] = val
        FofN[N]["idx_a_min"] = idx
    colors = sns.color_palette("bright",len(Nset))
    fig, axs=plt.subplots(3)
    fig.set_figheight(8)
    fig.set_figwidth(5)
    for ii, mode in enumerate(["K'splay","K'bend","K'twist"]):
        for iN, N in enumerate(list(Nset)[::-1]):
            iN=len(Nset)-1-iN
            alpha=1.0
            if type(N) != type(0.1):
                continue
            if N<0.3:
                continue
            a_values = FofN[N]["a2lpAf"]
            idx = FofN[N]["idx_a_min"]
            if mode=="K'splay":
                if N>0.9:
                    label=str(N)
                else:
                    label=None
            elif mode=="K'twist":
                if N<4.0:
                    label=str(N)
                else:
                    label=None
            else:
                label=None
            axs[ii].plot(a_values[idx:], FofN[N][mode][idx:]*N, 
                     color=colors[iN], alpha=alpha,
                     label=label)
        if log_gamma:
            axs[ii].xscale("log")
        if logy:
            axs[ii].yscale("log")
        #axs[ii].set_xlabel("a(2lpAf)")
        axs[ii].set_ylabel("K'*N")
        axs[ii].set_xlim([17,107])
        if mode=="K'splay":
            axs[ii].set_ylim([0,45])
            axs[ii].legend(loc=1, bbox_to_anchor=(0.97,0.97))
        elif mode=="K'bend":
            axs[ii].set_ylim([0,0.8])
        elif mode=="K'twist":
            axs[ii].set_ylim([0,0.04])
            axs[ii].legend(loc=4)
        #axs[ii].set_title(mode)
    plt.tight_layout()
    fig.savefig("Subplots.pdf")
    plt.show()

def find_gamma(N, aset = None, gammaN = True, gamma_TOL=10**-6, a_TOL=10**-6):
    """Frank Elastic data for given N and a

    Args:
        N (float): Number of Kuhn length
        aset (itrable): aLAf or a2lpAf values
        gammaN (bool): If ture use aLAf, else use a2lpAf, amoung other things
    """
    if gammaN:
        if aset is None:
            aset = [20, 30, 40, 50, 60]
        if N==0:
            #in this case get_a will take gammaN
            gamma_min=2.0
            gamma_max=80.0
        else:
            gamma_min = 2.0/N
            gamma_max = 80.0/N
        astring = "aLAf"
    else:
        if aset is None:
            aset = [20, 25, 30, 35, 40]
        gamma_min = 5.0
        gamma_max = 25.0
        astring="a2lpAf"
    gamma_max2 = gamma_max
    
    pairs = {}
    pairs[gamma_max] = get_a(gamma_max, N)[astring]
    pairs[gamma_min] = get_a(gamma_min, N)[astring]


    while True:
        if gamma_max-gamma_min<gamma_TOL:
            break
        if abs(pairs[gamma_max]-pairs[gamma_min])<a_TOL:
            break
    
        gamma_middle = (gamma_max+gamma_min)*0.5
        pairs[gamma_middle] = get_a(gamma_middle, N)[astring]
        #print("gamma %f, a %f"%(gamma_middle,pairs[gamma_middle]))
        if pairs[gamma_max]>pairs[gamma_min]:
            gamma_max = gamma_middle
        else:
            gamma_min = gamma_middle
    
    out={"a_min":pairs[gamma_max], "gamma_at_amin":gamma_max}
    out["aset"]=aset
    out["frank_data"] = []
    out["gammas"] = []
    out["N"]=N

    for a_desire in aset:
        #print("a desire %f"%(a_desire))
        gamma_max = gamma_max2
        gamma_min = out["gamma_at_amin"]
    
        for (gamma, aa) in pairs.items():
            if gamma<=gamma_min:
                continue
            if gamma>=gamma_max:
                continue
            if aa>a_desire:
                gamma_max = gamma
            elif aa<a_desire:
                gamma_min = gamma
        
        while True:
            if a_desire < out["a_min"]:
                gamma_best=None
                break
            a_at_min = pairs[gamma_min]
            a_at_max = pairs[gamma_max]
            if gamma_max-gamma_min<0 or a_at_max-a_at_min<0:
                print("N"+str(N))
                pdb.set_trace()
                raise ValueError("max<min")
            if gamma_max-gamma_min<gamma_TOL or a_at_max-a_at_min<a_TOL:
                gamma_best=(gamma_max+gamma_min)*0.5
                break
            
            gamma_middle = gamma_min + (gamma_max-gamma_min)\
                                       *(a_desire-a_at_min)\
                                       /(a_at_max-a_at_min)

            a_middle = get_a(gamma_middle, N)[astring]
            #print("gamma %f, a %f"%(gamma_middle, a_middle))
            pairs[gamma_middle] = a_middle
            if abs(a_middle-a_desire)<a_TOL:
                gamma_best = gamma_middle
                break
            if a_middle<a_desire:
                gamma_min = gamma_middle
            else:
                gamma_max = gamma_middle 

        out["gammas"].append(gamma_best)
        if gamma_best is None:
            #print("gamma_best = None")
            out["frank_data"].append(None)
        else:
            #print("gamma_best %f"%(gamma_best))
            out["frank_data"].append(get_Frank_values(gamma_best, N))

    return out

def get_Nsweep_data(saveas="Nsweep.p", MultiProcess=True, nthreads=25, Ns=None):
    """Calculate Frank Elastic Constants for sweep over N


    Args:
        saveas= Where to save results.
        nthreads (int): Number of threads to use
        MultiProcess (bool): Use many threads
        Ns (iterable): N values
    """
    if Ns is None:
        npts = 80
        #Ns = list( 3.1*np.logspace(-2,0,npts) )
        Ns = list( np.linspace(0.02,3.0,npts) )
    data = {"Ns":Ns}
    
    if MultiProcess:
        #if __name__ == '__main__':
        with Pool(nthreads) as my_pool:
            results = my_pool.map(find_gamma, Ns)
         
        data["results"] = results
        for ii, N in enumerate(Ns):
            data[N] = results[ii]
    else:
        for N in Ns:
            print("----- Working on N=%f ----- "%(N))
            data[N] = find_gamma(N)

    pickle.dump(data, open(saveas,"wb"))


def plotNsweep(load):
    """Plot results from get_Nsweep_data

    Args:
        load (string): File to load from.
    """
    #data = pickle.load(open("Nsweep2.p","rb"))
    data = pickle.load(open(load,"rb"))
    dataRR = find_gamma(0.0)["frank_data"]

    Ns = np.array(data["Ns"])
    npts = len(Ns)
    bend = np.zeros(npts)
    twist = np.zeros(npts)
    splay = np.zeros(npts)
    aset = data["results"][0]["aset"]
    colors = sns.color_palette("deep",len(aset))
    for jj, aa in enumerate(aset[::-1]):
        ia=len(aset)-1-jj
        color = colors[ia]
        if False:
            if ia == 0:
                labels=["bend/twist", "splay/twist", "bend/splay"]
            else:
                labels=[None, None, None]
        else:
            labels=[str(aa), None, None]

        for ii, N in enumerate(Ns):
            if data["results"][ii]["frank_data"][ia] is None:
                bend[ii]=np.nan
                twist[ii]=np.nan
                splay[ii]=np.nan
                continue
            bend[ii] = data["results"][ii]["frank_data"][ia]["K'bend"]
            twist[ii] = data["results"][ii]["frank_data"][ia]["K'twist"]
            splay[ii] = data["results"][ii]["frank_data"][ia]["K'splay"]

        linestyle = ["-","--",":"]
        #plt.plot(Ns, bend/twist, color=color, linestyle=linestyle[0], label=labels[0])
        #plt.plot(Ns, splay/twist, color=color, linestyle=linestyle[1], label=labels[1])
        #plt.plot(Ns, bend/splay, color=color, linestyle=linestyle[2], label=labels[2])

        #plt.plot(0.0, dataRR[ia]["K'bend"]/dataRR[ia]["K'twist"], "o", color=color)
        #plt.plot(0.0, dataRR[ia]["K'splay"]/dataRR[ia]["K'twist"], "o", color=color)
        xx = np.insert(Ns, 0, 0.0, axis=0) 
        yy = np.insert(bend/twist, 0,
                       dataRR[ia]["K'bend"]/dataRR[ia]["K'twist"], axis=0) 
        plt.plot(xx, yy, color=color, linestyle=linestyle[0], label=labels[0])
        yy = np.insert(splay/twist, 0,
                       dataRR[ia]["K'splay"]/dataRR[ia]["K'twist"], axis=0) 
        plt.plot(xx, yy, color=color, linestyle=linestyle[1], label=labels[1])
    plt.legend()
    #plt.xscale("log")
    plt.yscale("log")
    plt.xlim([0.0,2.0])
    plt.ylim([3.0,100.0])
    plt.plot()
    plt.ylabel("Ratio of Elastic Constants")
    plt.xlabel("N")
    plt.savefig("Nsweep.pdf")
    plt.show()





"""Module to plot polymers, cylinders, and spheres."""
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
import pandas as pd
from .utils.utils import well_behaved_decorator, make_decorator_factory_args_optional
from scipy import stats
from matplotlib.widgets import Slider, Button, RadioButtons
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import Qt


def make_standard_axes(is_3D=False):
    if is_3D:
        return plt.figure().add_subplot(111, projection='3d')
    else:
        return plt.figure().add_subplot(111)

@well_behaved_decorator(has_params=True)
@make_decorator_factory_args_optional
def make_axes_if_blank(is_3D=False):
    """Decorator for plotting function that take an "axes" kwarg, which will
    create an axes for them to use if one isn't passed to them."""
    def wrap(plot_function):
        """make_axes_if_blank's function that wraps plot_function."""
        def wrapped_f(*args, **kwargs):
            """The plot_function wrapped by make_axes_if_blank's wrap."""
            if 'axes' not in kwargs:
                kwargs['axes'] = make_standard_axes(is_3D)
            return plot_function(*args, **kwargs)
        return wrapped_f # return decorated function
    return wrap

@make_axes_if_blank(is_3D=False)
def testplot2(x, y, **kwargs):
    """Testing2 make_axes_if_blank decorator."""
    plt.plot(x, y, axes=kwargs['axes'])

@make_axes_if_blank
def testplot3(x, y, **kwargs):
    """Testing3 make_axes_if_blank decorator."""
    plt.plot(x, y, axes=kwargs['axes'])

@make_axes_if_blank(is_3D=True)
def testplot4(x, y, z, **kwargs):
    """Testing4 make_axes_if_blank decorator."""
    plt.plot(x, y, axes=kwargs['axes'])

@make_axes_if_blank(is_3D=True)
def draw_sphere(x0, r, **kwargs):
    """Draw a 3D sphere with center x0 and radius r."""
    # how densely gridding should happen
    longitudes = kwargs.pop('longitudes', 20)
    longitude_count = longitudes*1j
    latitude_count = longitude_count/2
    u, v = np.mgrid[0:2*np.pi:longitude_count, 0:np.pi:latitude_count]
    x = x0[0] + r*np.cos(u)*np.sin(v)
    y = x0[1] + r*np.sin(u)*np.sin(v)
    z = x0[2] + r*np.cos(v)
    ax = kwargs.pop('axes', None)
    ax.plot_wireframe(x, y, z, **kwargs)
    return ax

@make_axes_if_blank(is_3D=False)
def locally_linear_fit(x, y, window_size=5, **kwargs):
    if window_size % 2 == 0:
        raise ValueError('window_size must be odd')
    if window_size < 3:
        raise ValueError('window_size must be at least 3, so that we have at'
                         ' least two points to fit at endpoints of array.')
    lenx = len(x)
    if lenx < 2:
        raise ValueError('Can\'t fit less than 2 points!')
    inc = int(window_size / 2)
    slope = np.zeros_like(x)
    for i in range(lenx):
        imin = np.max([0, i - inc])
        imax = np.min([lenx, i + inc + 1])
        slope[i], intcpt, r_val, p_val, std_err = stats.linregress(x[imin:imax], y[imin:imax])
    ax = kwargs.pop('axes', None)
    if ax is not None:
        ax.plot(x, slope, **kwargs)
    return ax, slope

class PolymerViewer(object):

    def __init__(self, r=None):
        # will keep track of which *index* in the time axis should be used to
        # draw the current scene
        self.t = 0
        self.r = r
        # define the actual axes layout
        self.fig = plt.figure()
        # area to plot polymers
        self.ax3d = plt.axes([0.05, 0.15, 0.9, 0.8], facecolor='w', projection='3d')
        # area to plot "slider" for selecting what time to plot
        self.ax = plt.axes([0.1, 0.025, 0.8, 0.05], facecolor='lightgoldenrodyellow')
        # # set up the QtGui panel and textbox
        # self.root = self.fig.canvas.manager.window
        # self.panel = QtWidgets.QWidget()
        # self.hbox = QtWidgets.QHBoxLayout(self.panel)
        # self.textbox = QtWidgets.QLineEdit(parent=self.panel)
        # self.hbox.addWidget(self.textbox)
        # self.panel.setLayout(self.hbox)
        # self.dock = QtWidgets.QDockWidget("Simulation Directory", self.root)
        # self.root.addDockWidget(Qt.BottomDockWidgetArea, self.dock)
        # self.dock.setWidget(self.panel)
        self.update_ax_limits()
        # slider to control what "time" is plotted, (0,1) is rescaled total
        # simulation time units
        self.slider_t = Slider(self.ax, 't', 0, 1, valinit=0)
        # handler to check if the "time" slider has been moved
        self.slider_t.on_changed(self.update_t)
        plt.show()

    @property
    def r(self):
        return self._r

    @r.setter
    def r(self, new_r):
        self._r = new_r
        self._num_time_points, self._num_beads, self._d = self._r.shape
        if hasattr(self, 'ax3d'):
            self.update_ax_limits()


    # should never be called before a Sim has been loaded successfully
    def update_drawing(self):
        num_polymers = 1
        if num_polymers != len(self.ax3d.lines):
            self.ax3d.lines = []
            for i in range(num_polymers):
                self.ax3d.plot(self.r[self.t,:,0],
                               self.r[self.t,:,1],
                               self.r[self.t,:,2])
        for i,line in enumerate(self.ax3d.lines):
            x, y, z = (self.r[self.t,:,0], self.r[self.t,:,1],
                       self.r[self.t,:,2])
            line.set_data(x, y)
            line.set_3d_properties(z)
        self.fig.canvas.draw_idle()

    def update_ax_limits(self):
        self.ax3d.set_xlim((np.nanmin(self.r[:,:,0]), np.nanmax(self.r[:,:,0])))
        self.ax3d.set_ylim((np.nanmin(self.r[:,:,1]), np.nanmax(self.r[:,:,1])))
        self.ax3d.set_zlim((np.nanmin(self.r[:,:,2]), np.nanmax(self.r[:,:,2])))

    def update_t(self, val):
        if self.r is None:
            return
        # calculate value on slider
        new_t = int(round(val*self._num_time_points))
        # bound it to 0, num_time_points-1
        new_t = max(0, min(self._num_time_points-1, new_t))
        self.t = new_t
        self.update_drawing()



""" This module "understands" the input format of wlcsim.exe """
from __future__ import print_function

import re
import os
from pathlib import Path
from enum import Enum

# def __init__(self, sim_dir):
#     input_dir = os.path.join(sim_dir, 'input')
#     input_file = os.path.join(input_dir, 'input')
#     if not os.path.isdir(sim_dir) \
#             or not os.path.isdir(input_dir) \
#             or not os.path.isfile(input_file):
#         raise IOError('Simulation directory ' + sim_dir + ' does not'
#                       ' contain input/input')

class InputFormat(Enum):
    ORIGINAL=1
    LENA=2
    DEFINES=3

renamer = {'COL_TYPE': 'COLLISIONDETECTIONTYPE', 'COLTYPE': 'COLLISIONDETECTIONTYPE', 'FPT_DIST': 'COLLISIONRADIUS',
           'INTON': 'INTERPBEADLENNARDJONES', 'N': 'NB', 'INDMAX': 'NUMSAVEPOINTS'}

def correct_param_name(name):
    """Takes messy param names from different generations of the simulator and
    converts them to their newest forms so that simulations from across
    different years can be tabulated together."""
    name = name.upper()
    if name not in renamer:
        return name
    else:
        return renamer[name]

def brown_to_codename(name, value):
    new_name = 'CODENAME'
    if int(value) == 1:
        return (new_name, 'bruno')
    else:
        return (new_name, 'brad')
# for each old param name that needs to go through correct_param_value, the
# function that will convert its value. recall all "values" are strings at this
# point
revalueer = {'BROWN': brown_to_codename}

def correct_param_value(name, value):
    """Some old param names also have new types. This takes messy param names
    from different generations of the simulator and converts their names and
    values to their newest forms so that simulations from across different
    years can be tabulated together."""
    if name in revalueer:
        return revalueer[name](name, value)
    else:
        return (name, value)

class ParsedInput(object):
    """Knows how to handle various input file types used by wlcsim simulator
    over the years, and transparently converts into new parameter naming
    conventions.

    input = ParsedInput(file_name)
    print(input.ordered_param_names) # see params in order defined
    print(input.ordered_param_values) # to see values
    input.write(outfile_name) # write clone of input file

    """

    def __init__(self, input_file=None, params=None):
        """Can be constructed from an input file or from params directly. If an
        input file is provided, params are just ignored if passed in."""
        self.params = {}
        self.ordered_param_names = []
        self.file_format = InputFormat.DEFINES
        self.input_file = Path(input_file)
        if input_file is not None:
            # if we get the sim dir or the input dir, resolve to the actual input file
            if not os.path.isfile(input_file) and os.path.isdir(input_file):
                input_file = os.path.join(input_file, 'input')
            if not os.path.isfile(input_file) and os.path.isdir(input_file):
                input_file = os.path.join(input_file, 'input')
            self.decide_input_format()
            self.parse_params_file()
        elif params is not None:
            for name, value in params.items():
                self.append_param(name, value)
        else:
            Warning('ParsedInput: no params or input file provided!')

    def __repr__(self):
        return str(self.params)

    @property
    def ordered_param_values(self):
        return [self.params[name] for name in self.ordered_param_names]

    def append_param(self, name, value):
        name = correct_param_name(name)
        name, value = correct_param_value(name, value)
        self.ordered_param_names.append(name)
        self.params.update({name: value})

    def write(self, output_file_name):
        """ writes out a valid input file for wlcsim.exe given parameters
            in the format returned by read_file """
        with open(output_file_name, 'w') as f:
            if self.file_format == InputFormat.ORIGINAL:
                # write the three garbage lines
                f.write('!\n!\n\n')
                for i,name in enumerate(self.ordered_param_names):
                    f.write('!-Record ' + str(i) + '\n!  ' + name + '\n')
                    f.write(str(self.params[name]) + '\n\n')
            elif self.file_format == InputFormat.LENA:
                for name in self.ordered_param_names:
                    f.write(str(name) + ' ' + str(self.params[name]) + '\n')
            elif self.file_format == InputFormat.DEFINES:
                for name in self.ordered_param_names:
                    f.write('#define WLC_P__' + str(name) + ' ' +
                            str(self.params[name]) + '\n')
            else:
                raise ValueError('wlcsim.input: attempt to print a ParsedInput'
                                 ' with unknown file_format.')

    def decide_input_format(self):
        """Decide between the two input formats we know of.
        Not too hard, since one uses Fortran-style comments, which we can look
        out for, and the other uses bash style comments. Further, the former
        specifies param names and values on separate lines, while the latter
        specifies them on the same line."""
        # first see if the file has the expected name for the defines file
        if self.input_file.name == 'defines.inc':
            self.input_format = InputFormat.DEFINES
            return
        # then see if there are any comment lines. if so, we immediately know
        # the file type
        with open(self.input_file) as f:
            for line in f:
                if not line:
                    continue
                elif re.match('\s*!', line):
                    self.input_format = InputFormat.ORIGINAL
                    return
                elif line[0] == '#':
                    self.input_format = InputFormat.LENA
                    return
                else:
                    continue
        # if there are no comments, then for now, it must in fact be Lena's
        # intput file type, otherwise, we would not be able to infer the param
        # names, since these are in comment lines in the original-type input
        # files
        self.input_format = InputFormat.LENA
        return


    def parse_params_file(self):
        """Parse and populate ParsedInput's params, ordered_param_names
        This parser currently understands three file formats:
        1) "ORIGINAL" is the input method understood by Andy's hardcoded
        input reader in the original code used the Spakowitz lab was
        founded.
        2) "LENA" is a spec using slightly more general input reader
        written by Elena Koslover while Andy's student.
        3) "DEFINES" is the format of the src/defines.inc file.
        """
        if self.input_format == InputFormat.ORIGINAL:
            self.parse_original_params_file()
        elif self.input_format == InputFormat.LENA:
            self.parse_lena_params_file()
        elif self.input_format == InputFormat.DEFINES:
            self.parse_defines_params_file()

    def parse_lena_params_file(self):
        """Lena-style input files have comment lines starting with a "#". Any
        other non-blank lines must be of the form
        "[identifier][whitespace][value]",
        where an identifier is of the form "[_a-zA-Z][_a-zA-Z0-9]*", and a
        value can be a boolean, float, int or string. They will always be
        stored as strings in the params dictionary for downstream parsing as
        needed.

        Identifiers, like fortran variables, are interpreted in a
        case-insensitive manner by the wlcsim program, and so will be store in
        all-caps within the ParsedInput to signify this."""
        name_value_re = re.compile('([_A-Za-z][_A-Za-z0-9]*)\s*(.*)\s*')
        with open(self.input_file) as f:
            for line in f:
                if not line or line[0] == '#':
                    continue
                match = name_value_re.match(line)
                if match is None:
                    continue
                name, value = match.groups()
                self.append_param(name, value)


    def parse_defines_params_file(self):
        """Parse file in the format of src/defines.inc. Each line begins with
        #define WLC_P__[PARAM_NAME] [_a-zA-Z0-9]
        where WLC_P__[A-Z]* is the parameter name and the piece after the space is the value of
        the parameter.

        TODO:test
        """
        name_value_re = re.compile('#define WLC_P__([_A-Z]*) ([_a-zA-Z0-9]*)')
        with open(self.input_file) as f:
            for line in f:
                if not line:
                    continue
                match = name_value_re.match(line)
                if match is None:
                    continue
                name, value = match.groups()
                self.append_param(name, value)


    def parse_original_params_file(self):
        """Original-style input files have three garbage lines at the top used
        for commenting then triplets of lines describing one variable each. The
        first line of the triplet is a counter, for ease of hardcoding input
        reader, the second line contains the variable name (which is not used
        by the input reader) in order to make parsing possible outside of the
        ahrdcoded wlcsim input reader, and the third line contains the value of
        the parameter itself. The first two lines of each triplet have a fixed
        form that we use to extract the record numbers and parameter names, but
        these forms are not used by wlcsim itself, which ignores these lnies
        completely. Thus, it is possible that the user specified a particular
        name for a parameter but that name does not match what wlcsim
        interpreted it as, since wlcsim simply uses the order of the parameters
        in this file to determine their identities."""
        record_re = re.compile('!-Record (\d+)')
        name_re = re.compile('! *([_A-Za-z0-9]+)')
        contains_period_re = re.compile('\.')
        next_line_is_val = False
        with open(self.input_file) as f:
            # first three lines are garbage
            for i in range(3):
                f.readline()
            for line in f:
                if next_line_is_val:
                    if contains_period_re.search(line):
                        value = float(line.strip())
                    else:
                        value = int(line.strip())
                    self.append_param(name, value)
                    name = None
                    record_number = None
                    value = None
                    next_line_is_val = False
                record_match = record_re.search(line)
                if record_match:
                    record_number = int(record_match.groups()[0])
                    continue
                name_match = name_re.search(line)
                if name_match:
                    name = name_match.groups()[0]
                    name = name.upper()
                    next_line_is_val = True
                    continue



"""
Force fields used in our BD integrators.

Our BD code can be summarized as typically looking like:

.. code-block:: python

    X = init_function()
    for t_i in requested_times:
        X = integrator_step(f_all, X, prev_t, t_i)
        prev_t = t_i

where ``f_all`` takes ``X`` (size ``N_beads`` by ``N_dimensions``) and ``t``
and returns the total force on each of the beads at time ``t``.

This module holds all the functions that can be used to construct ``f_all``.
Broadly, they are named based on the source of the force:

- ``f_elas_*`` are elastic forces, they define the physics of the polymer being
  simulated. Linear, ring, and different types of network polymers have their
  own independent implementations.
- ``f_conf_*`` are confinement forces.
- ``f_tether_*`` tether individual beads of the polymer to an external
  structure.
- Others may be added in the future.

Notes
-----
Because they are in the "innermost loop", it is important that these functions
are just-in-time compile-able when possible.
"""
import sys
import inspect

import numpy as np
from numba import jit


# Append this to all f_tether_* docstrings to make it clear how they should
# actually be used.
extra_tether_docs = """

    Notes
    -----
    The force computed here is directed at the confinement from the inside
    only. To properly "tether" a bead to a surface, use the sum of the relevant
    ``f_conf + f_tether`` functions.
"""


@jit(nopython=True)
def f_conf_ellipse(x0, Aex, rx, ry, rz):
    """Compute soft (cubic) force due to elliptical confinement."""
    N, _ = x0.shape
    f = np.zeros(x0.shape)
    for j in range(N):
        conf = x0[j, 0]**2/rx**2 + x0[j, 1]**2/ry**2 + x0[j, 2]**2/rz**2
        if conf > 1:
            conf_u = np.array([
                -x0[j, 0]/rx**2, -x0[j, 1]/ry**2, -x0[j, 2]/rz**2
            ])
            conf_u = conf_u/np.linalg.norm(conf_u)
            # Steph's confinement from
            # https://journals.aps.org/pre/abstract/10.1103/PhysRevE.82.011913
            f[j] += Aex*conf_u*np.power(np.sqrt(conf) - 1, 3)
    return f


@jit(nopython=True)
def f_tether_ellipse(x0, Aex, rx, ry, rz):
    """Compute soft (cubic) force towards an elliptical confinement."""
    N, _ = x0.shape
    f = np.zeros(x0.shape)
    for j in range(N):
        conf = x0[j, 0]**2/rx**2 + x0[j, 1]**2/ry**2 + x0[j, 2]**2/rz**2
        if conf < 1:
            conf_u = np.array([
                -x0[j, 0]/rx**2, -x0[j, 1]/ry**2, -x0[j, 2]/rz**2
            ])
            conf_u = conf_u/np.linalg.norm(conf_u)
            # Steph's confinement from
            # https://journals.aps.org/pre/abstract/10.1103/PhysRevE.82.011913
            f[j] += Aex*conf_u*np.power(np.sqrt(conf) - 1, 3)
    return f


@jit(nopython=True)
def f_elas_linear_rouse(x0, k_over_xi):
    """Compute spring forces on single, linear rouse polymer."""
    N, _ = x0.shape
    f = np.zeros(x0.shape)
    for j in range(1, N):
        for n in range(3):
            f[j, n] += -k_over_xi*(x0[j, n] - x0[j-1, n])
            f[j-1, n] += -k_over_xi*(x0[j-1, n] - x0[j, n])
    return f


@jit(nopython=True)
def f_elas_homolog_rouse(x0, N, loop_list, k_over_xi):
    r"""
    Compute spring forces on two "homologously linked linear rouse polymers.

    The two polymers are (homologously) at beads specified by loop_list. While
    each polymer is of length N beads, the beads that are hooked together move
    together identically, so you have `x0.shape[0] == 2*N - len(loop_list)`

    We assume that, per the spec in `~.homolog_points_to_loops_list`, a polymer
    of the shape

    .. code-block:: text

        ++   ^^^   '''''''   ""   =
          \ /   \ /       \ /  \ /
           .     .         .    .
          / \   / \       / \  / \
        --   ---   -------   --   -

    will be laid out in memory like

    .. code-block:: text

        --.---.-------.--.-++^^^'''''''""=

    If we call the lower polymer "polymer 1", and all the beads on the top row
    "polymer 2", then it makes sense to first compute the spring forces between
    adjacent beads in polymer 1 as if there was no "polymer 2". Then, we can
    compute all the forces between beads in "polymer 2" and the "." beads.
    Finally, for each "loop" (represented as symbols of different types above)
    in polymer 2, we can compute the usual linear forces along the polymer.

    """
    f = np.zeros(x0.shape)
    num_loops, _ = loop_list.shape
    # first get forces linearly along "first" polymer
    for j in range(1, N):
        f[j] += -k_over_xi*(x0[j] - x0[j-1])
        f[j-1] += -k_over_xi*(x0[j-1] - x0[j])
    # in what follows, "loop" means the stretches between homologous
    # connections, including from the homologous connection to the end of the
    # polymer
    for i in range(num_loops):
        # k1l, k1r denote the connected beads, so the array contains a "loop"
        # representing the beads of the second chain from k1l+1 to k1r-1
        # at array locations [k2l, k2r].
        k1l, k1r, k2l, k2r = loop_list[i]
        # k1l == -1 is flag for "end" loop (free end)
        # k2l > k2r => no beads in "second chain"
        if k1l >= 0 and k2l <= k2r:
            f[k1l] += -k_over_xi*(x0[k1l] - x0[k2l])
            f[k2l] += -k_over_xi*(x0[k2l] - x0[k1l])
        # k1r == N is flag for "end" loop (free end)
        if k1r < N and k2l <= k2r:
            f[k1r] += -k_over_xi*(x0[k1r] - x0[k2r])
            f[k2r] += -k_over_xi*(x0[k2r] - x0[k1r])
        N_loop = k2r - k2l + 1
        # analagous to loop above, since end-bead forces already done
        for k in range(1, N_loop):
            j = k2l+k
            f[j] += -k_over_xi*(x0[j] - x0[j-1])
            f[j-1] += -k_over_xi*(x0[j-1] - x0[j])
    return f


@jit(nopython=True)
def f_elas_linear_sswlc(x, u, e_b, gam, e_par, e_perp, eta, xi_u, xi_r):
    """Compute spring forces and torques on each bead of dsswlc."""
    N, _ = x.shape
    f = np.zeros(x.shape)
    t = np.zeros(x.shape)
    for i in range(0, N - 1):
        dx = x[i+1] - x[i]
        dx_par = dx @ u[i]
        dx_perp = dx - dx_par*u[i]
        cos_u1_u2 = u[i+1]@u[i]

        Gi = u[i+1] - cos_u1_u2*u[i] - eta*dx_perp
        Fi = -eta*e_b*Gi + e_par*(dx_par - gam)*u[i] + e_perp*dx_perp
        f[i] += Fi
        f[i + 1] -= Fi

        Gi = (u[i+1] - u[i]) - eta*dx_perp
        t[i] += e_b*Gi - eta*e_b*dx_par*Gi + eta*e_b*(1 - cos_u1_u2)*dx \
            - e_par*(dx_par - gam)*dx + e_perp*dx_par*dx_perp
        t[i+1] -= e_b*Gi
    return f, t


for name, obj in inspect.getmembers(sys.modules[__name__]):
    if not inspect.isfunction(obj) or not name.startswith('f_tether'):
        continue
    obj.__doc__ += extra_tether_docs



from numpy import sqrt
import numpy as np
#from util import sphinx_compat_jit as jit
from numba import jit

ORDER_L=50

@jit
def alpha(l,m):
    return sqrt((3*(l-m)*(l+m))/(4*np.pi*(2*l-1)*(2*l+1)))
    
@jit
def alpha_plus(l,m):
    return sqrt((3*(l+m)*(l+m+1))/(8*np.pi*(2*l-1)*(2*l+1)))

@jit
def Alm(l,m):
    return alpha(l,m)*alpha(l-1,m)/alpha(2,0)

@jit
def Blm(l,m):
    return (alpha(l+1,m)*alpha(l+1,m) 
            - alpha(1,0)*sqrt(0.25/np.pi) +
            alpha(l,m)*alpha(l,m))/alpha(2,0)

@jit
def PgammaB_vec(m, p, gamma):
    """ P - \gamma \\beta
        
    Returns:
        vector with index ell
    """
    PgammaB = np.zeros(ORDER_L, np.complex128)
    for ell in range(abs(m),ORDER_L):
        PgammaB[ell] = p+ell*(ell+1) - gamma*Blm(ell,m)
    return PgammaB

@jit
def Alm_vec(m):
    Am_vec = np.zeros(ORDER_L, np.complex128)
    for ell in range(abs(m)+2,ORDER_L):
        Am_vec[ell] = Alm(ell,m)
    return Am_vec

@jit 
def Wplus_vec(m, gamma, p, Am, PgammaB):
    Wplus = np.zeros(ORDER_L, np.complex128)
    for ell in (ORDER_L-1,ORDER_L-2):
        Wplus[ell] = 1.0/PgammaB[ell]
    for ell in range(ORDER_L-3,abs(m)-1,-1):
        Wplus[ell] = 1.0/(PgammaB[ell] - (gamma*Am[ell+2])**2*Wplus[ell+2])
    return Wplus

@jit 
def Wminus_vec(m, gamma, p, Am, PgammaB):
    Wminus = np.zeros(ORDER_L, np.complex128)
    for ell in (abs(m),abs(m)+1):
        Wminus[ell] = 1.0/PgammaB[ell]
    for ell in range(abs(m)+2,ORDER_L):
        Wminus[ell] = 1.0/(PgammaB[ell] - (gamma*Am[ell])**2*Wminus[ell-2])
    return Wminus


@jit
def Gmll_matrix(Wplus, Wminus, Am, PgammaB, gamma, m):
    """"Matrox of propagators between starting and ending l value.

    Args:
        Wplus (numpy array): Result of Wplus_vec for same m, p
        Wminus (numpy array): Reult of Wminus_vec for same m, p
        Am (numpy array): Result of Am_vec for same m
        PgammaB (numpy array): Result of PgammaB_vec for same m, p, gamma
        gamma (float): alignment strength, in kT's per Kuhn length
        m (int): z component of agular momentum quantum number

    Returns:
        An ORDER_L x ORDER_L numpy matrix with propagators that use Maier-Saupe
        steps to get from l0 to lf.
    """
    Wpm = np.zeros(ORDER_L,np.complex128)
    absm = abs(m)
    Wpm[absm:] = (Wplus[absm:]*Wminus[absm:])\
                 /(Wminus[absm:] 
                   - PgammaB[absm:]*Wplus[absm:]*Wminus[absm:] + Wplus[absm:])
    Gmll = np.zeros((ORDER_L,ORDER_L), np.complex128)
    for ell in range(abs(m),ORDER_L):
        Gmll[ell,ell] = Wpm[ell]
        for lf in range(ell+2,ORDER_L,2):
            Gmll[ell, lf] = Gmll[ell, lf-2]*Wplus[lf]*Am[lf]*gamma
            Gmll[lf, ell] = Gmll[ell, lf] # Must be symmetric
        # the following loop is an alturnative the using the symmetric propory
        #for lf in range(ell-2,-1,-2):
        #    Gmll[ell, lf] = Gmll[ell, lf+2]*Wminus[lf]*Am[lf+2]*gamma
    return Gmll

def precalculate_data(p, gamma, m_values=[0]):
    """Precalculate W_plus, W_minus, W_pm, and G_m_ll

    Args:
        p (complex): laplace conjugate of path length
        gamma (real): aligning l=2 (Maier-Saupe) field strength
        m_values (list): list of integer m values to precalculate for
    """
    
    Wps = {}
    Wms = {}
    Gmlls = {}
    for m in m_values:
        Am = Alm_vec(m)
        PgammaB = PgammaB_vec(m, p, gamma)
        Wplus = Wplus_vec(m, gamma, p, Am, PgammaB)
        Wminus = Wminus_vec(m, gamma, p, Am, PgammaB)
        Gmll = Gmll_matrix(Wplus, Wminus, Am, PgammaB, gamma, m)

        Wps[m]=Wplus
        Wms[m]=Wminus
        Gmlls[m] = Gmll

    return {"Wplus":Wps, "Wminus":Wms, "Gmll":Gmlls, "ms":m_values, "p":p,
            "gamma":gamma}




""""Simulation" of two yeast sister chromatids forming a synaptonemal
complex"""
import numpy as np
from numba import jit

from ..analytical.rouse import gaussian_Ploop

UNLOOP = 0
PAIRED = 1
JUNC_LEFT_END = 2
JUNC_RIGHT_END = 4
JUNC_BOTH = 6

@jit(nopython=True)
def Ploop_given_paired(n, starts, ends, b, dN, a):
    r"""Get looping probability for a "homologously paired" pair of Rouse
    chains.

    Parameters
    ----------
    n : int
        number of beads in "state"
    starts/ends : np.array<int>
        output of get_terminatable_loops
    b : float
        Kuhn length of linear chain
    dN : float
        number of kuhn length between each pair of beads
    a : float
        capture radius of looping probability calculation

    Returns
    -------
    ploop : (n,) np.array<float>
        probability that a site will be looped given conformation specified.
        sites not inside the starts/ends dilineators are given 0 probability of
        looping

    Notes
    -----
    Uses the fact that two loci on a ring polymer (Kuhn length `b` and
    separations `N`, `L-N`) can be thought of as being connected by an
    effective linear polymer.

    We simply call $k_1 \propto 1/Nb^2$ and $k_2 \propto 1/(L-N)b^2$ to be the
    spring constants of the two arms of the loop, then they form one combined
    spring additively, with $k_\text{tot} = k_1 + k_2$, meaning that any
    linaer polymer of length $\hat{N}$ and Kuhn length $\hat{b}$ will have the
    same looping probability at its ends as our two loci, as long as

    $$ \hat{N}\hat{b}^2 = \frac{1}{1/Nb^2 + 1/(L-N)b^2}. $$
    """
    ploop = np.zeros((n,))
    if len(starts) == 0:
        return ploop
    # for each loop site, we compute the looping probabilities to its left
    for i, si in enumerate(starts):
        ei = ends[i]
        # the current "loop" is the whole polymer
        if si == -1 and ei == n:
            #TODO: use conf gaus green's function to get actual "free" ploop
            pass # assume floor ploop will be set outside of here
        # the current "loop" is the free tails at the start of the polymer
        elif si == -1:
            ploop[:ei] = gaussian_Ploop(a, 2*dN*(ei - np.arange(ei)), b)
        # the current "loop" is the free tails at the end of the polymer
        elif ei == n:
            tail_len = (n - 1) - si
            ploop[si+1:] = gaussian_Ploop(a,
                    2*dN*(tail_len - np.arange(tail_len)[::-1]), b)
        # the current "loop" is in between si and ei, with total length
        else:
            L = 2*dN*(ei - si)
            # length of the one arm of the loop connecting each homolog pair
            N = 2*dN*np.arange(1, ei - si)
            Nb2_hat = 1 / ( 1/N/b/b + 1/(L-N)/b/b )
            ploop[si+1:ei] = gaussian_Ploop(a, Nb2_hat, 1)
    return ploop

@jit(nopython=True)
def get_terminatable_loops(state):
    """Take a state array and get the ends of the loops that have yet to
    terminate, so that we can count how many termination reactions are possible
    given the current state."""
    starts = []
    ends = []
    num_beads = len(state)
    # keep track of if we're at loci that have been "zippered" into the
    # synaptonemal complex yet (i.e. those that rest between related "junction"
    # sites (in a string that matches the regex "[26]0*[46]"))
    in_junction = state[0] == JUNC_LEFT_END
    if not in_junction:
        # check if the end of the polymer is "free" or in a loop/junction
        in_loop = state[0] == UNLOOP
        # special flag for if the end of the polymer is not looped, treat as though
        # an imaginary "-1" bead is paired. otherwise ensure we start with one more
        # starts than ends
        starts.append(-1 if in_loop else 0) # else PAIRED
    for i, si in enumerate(state):
        if i == 0:
            continue
        elif not in_junction and si == PAIRED:
            ends.append(i)
            starts.append(i) # now N_ends = N_starts - 1
        elif not in_junction and si == JUNC_LEFT_END:
            ends.append(i) # now N_ends == N_starts
            in_junction = True
        elif not in_junction and (si == JUNC_BOTH or si == JUNC_RIGHT_END):
            raise ValueError("How did we end up hitting right end of junction first?")
        elif not in_junction and si == UNLOOP:
            continue
        elif in_junction and si == PAIRED:
            raise ValueError("No loops should form in between junction points.")
        elif in_junction and si == JUNC_LEFT_END:
            raise ValueError("How did we hit left end of junction from inside junction?")
        elif in_junction and si == JUNC_BOTH:
            continue
        elif in_junction and si == JUNC_RIGHT_END:
            starts.append(i) # now N_ends = N_starts - 1
            in_junction = False
        elif in_junction and si == UNLOOP:
            continue
    if si == PAIRED or si == JUNC_RIGHT_END:
        starts.pop() # unecessarily added a "starts"
    elif not in_junction and si == UNLOOP:
        ends.append(num_beads)
    elif in_junction and si == UNLOOP:
        raise ValueError("What is the junction with if the end bead is UNLOOP?")
    # other impossibilities
    # checked above: (not in_junction and si == JUNC_RIGHT_END) or (in_junction
    # and si == PAIRED)
    elif si == JUNC_LEFT_END or si == JUNC_BOTH:
        raise ValueError("Rightmost bead cannot be left end of junction.")
    return np.array(starts), np.array(ends)

def test_get_terminatable_loops():
    starts, ends = get_terminatable_loops(np.array([0, 0, 1, 1, 0, 1, 2, 2, 0, 0, 1]))
    act_starts = np.array([-1, 2, 3, 5, 7])
    act_ends = np.array([2, 3, 5, 6, 10])
    assert(np.all(act_starts == starts))
    assert(np.all(act_ends == ends))
    starts, ends = get_terminatable_loops(np.array([1, 0, 1, 1, 0, 1, 2, 2, 0, 0, 0]))
    act_starts = np.array([0, 2, 3, 5, 7])
    act_ends = np.array([2, 3, 5, 6, 11])
    assert(np.all(act_starts == starts))
    assert(np.all(act_ends == ends))

# @jit(nopython=True)
def chromatid_gillespie(init_paired, b, N, k_h, k_u, a,
                        tend=np.inf, target_completion=1, k_t=None, R=np.inf,
                        save_times=None):
    """Do Gillespie sim of homolog pairing process.

    Parameters
    ----------
    init_paired : array_like of bool
        Which loci should be initially paired. length of array determines
        number of "beads". for now, at least one locus should be paired.
    b : float
        Kuhn length
    N : float
        total length of polymer in number of Kuhn lengths
    k_h : float
        rate of stable homolog junctions given loop exists
    a : float
        capture radius inside which stable homolog junction can form
    tend : float, optional
        what the cutoff time for the simulation should be. `np.inf` by default,
        where no limit implies `target_completion` will be the only termination
        condition.
    target_completion : float, optional
        $\in[0,1]$ what fraction of beads should be marked as "completed"
        before the simulation ends. since there is not way to undo
        "synaptonemal formation" reactions, it does not make sense to continue
        the simulation after this fraction reaches 1, as nothing can happen. 1
        by default.
    k_t : float, optional
        rate of synaptonemal junction formation (i.e. "termination" reaction),
        equal to k_h by default
    R : float, optional
        the radius of the spherical confinement. only used if there are no
        initial pairings, in order to have a non-zero probability that the
        first reaction occurs. once the first reaction occurs, it is assumed
        that this "random collision" reaction will never happen again

    Notes
    -----
    `state` is an int array of 0's, 1's, and 2's. 0 means unpaired, 1 means
    paired, and 2 means the locus has been sequestered into the synaptonemal
    complex
    """
    state = np.array(init_paired).astype(int)
    if k_t is None:
        k_t = k_h
    if a <= 0 or tend < 0 or k_h < 0 or k_t < 0 or N <= 0 or b <= 0 \
       or R <= 0 or np.any(state < 0) or np.any(state > 2):
        raise ValueError("One of the inputs makes no sense.")
    if np.isinf(R) and k_u > 0:
        raise ValueError("You probably don't want unbinding reactions if " \
                         "there's no way for a totally unpaired polymer to " \
                         "find a loop (aka R=np.inf")
    if save_times is None:
        save_times = [0, np.inf]
    num_beads = len(state)
    dN = N/(num_beads-1)
    min_ploop = a**3 / R**3 # fraction of volume occupied by interaction sphere
    t = 0
    save_i = 0
    save_states = []
    # the only loci that can "pair" are the ones that are inside of loops
    # that have not undergone synaptonemal junction formation, so get the
    # boundaries of those loops
    term_starts, term_ends = get_terminatable_loops(state)
    while t < tend and len(term_starts) > 0:
        while t >= save_times[save_i]:
            save_states.append(state.copy())
            save_i += 1
        if len(term_ends) > 1:
            # the synaptonemal junction formation is also a looping process, simply
            # of two loci on exactly opposite ends of the "loop", which is the same
            # as being on a linear polymer with half that separation
            ploop = gaussian_Ploop(a, dN*(term_ends - term_starts)/2, b)
            rates_per_term = k_t*np.maximum(min_ploop, ploop)
            # deal with special "end" cases, corresponding to free ends looping
            left_ploop = gaussian_Ploop(a, (num_beads-1) - term_ends[term_starts == -1], b)
            rates_per_term[term_starts == -1] = k_t*np.maximum(min_ploop, left_ploop)
            right_ploop = gaussian_Ploop(a, term_starts[term_ends == num_beads], b)
            rates_per_term[term_ends == num_beads] = k_t*np.maximum(min_ploop, right_ploop)
        else: # special case of full polymer unlooped
            ploop = gaussian_Ploop(a, dN*num_beads, b)
            rates_per_term = k_t*np.maximum(min_ploop, ploop)
        # make array to index into for gillespie reaction "choice"
        term_cumulat = np.cumsum(rates_per_term)
        term_tot = term_cumulat[-1]
        # the other two reactions are simple looping and unlooping
        # only non-terminated pairs can unloop
        rates_per_site = k_u*(state==1).astype(float)
        # get looping-weighted reaction rates for all loci that can loop
        rates_per_site += k_h*(state==0)*Ploop_given_paired(num_beads, term_starts, term_ends, b, dN, a)
        site_cumulat = np.cumsum(rates_per_site)
        site_tot = site_cumulat[-1]
        # first select which reaction to do
        if np.random.rand() < term_tot / (term_tot + site_tot):
            # choose which of the loops will close
            ti = np.searchsorted(term_cumulat, term_tot*np.random.rand())
            # a termination reaction marks from locus to the next looped site
            # as "done", but make sure to not write out of bounds for two
            # special cases
            si = max(term_starts[ti], 0)
            state[si] = (state[si] & ~PAIRED) | JUNC_LEFT_END
            ei = min(term_ends[ti], num_beads-1)
            state[ei] = (state[ei] & ~PAIRED) | JUNC_RIGHT_END
        else:
            # choose which of the sites will react
            reacted_site = np.searchsorted(site_cumulat, site_tot*np.random.rand())
            # a looping reaction just switches the state of that locus
            state[reacted_site] = 1 - state[reacted_site]
        # even though we selected the reaction first, we just still step time
        # forward using the total cumulative rate for all reactions, since any
        # of them were in fact "possible"
        dt = np.random.exponential(scale=1/(site_tot+term_tot))
        # update time and our loop list
        t += dt
        term_starts, term_ends = get_terminatable_loops(state)
    if np.inf in save_times:
        save_states.append(state.copy())
    return save_states



import os
import glob
import shutil

def fix_coltimes(coldir, newdir, resume=None):
    dirs = glob.glob(coldir + '/*/core.*')
    os.makedirs(newdir, mode=0o755, exist_ok=True)
    if resume:
        have_reached_resume = False
    else:
        have_reached_resume = True
    for sim in dirs:
        if not have_reached_resume:
            if sim == resume:
                have_reached_resume = True
            else:
                continue
        hostname = sim
        hostname = hostname[hostname.find('/')+1:]
        hostname = hostname[:hostname.find('/')]
        run_base = os.path.join(newdir, hostname)
        run_num = 0
        while True:
            try:
                run_dir = run_base + '.' + str(run_num)
                os.makedirs(run_dir, mode=0o755)
                break
            except OSError:
                run_num = run_num + 1
        shutil.copytree(os.path.join(sim, 'input'), os.path.join(run_dir, 'input'))
        cpfile = lambda f: shutil.copyfile(os.path.join(sim, f),
                                           os.path.join(run_dir, f))
        cpfile('wlcsim.exe')
        cpfile('run_parameter.pl')
        cpfile('runsim.sh')
        for i,finished_sims in enumerate(glob.glob(sim + '/savedata/*')):
            finished_dir = run_base + '.' + str(run_num + i + 1)
            shutil.copytree(run_dir, finished_dir)
            shutil.copytree(finished_sims, os.path.join(finished_dir, 'data'))
        shutil.copytree(os.path.join(sim, 'data'), os.path.join(run_dir, 'data'))

def fix_old0(old0_dir, newdir):
    dirs = glob.glob(old0_dir + '/tower*/par-run-dir.*/run.*')
    os.makedirs(newdir, mode=0o755, exist_ok=True)
    for sim in dirs:
        if not os.path.isdir(os.path.join(sim, 'input')):
# old.0/tower0/par-run-dir.1/run.25 is a special snowflake
# here is where we would use him if we wanted to, but doesn't seem
# like anything important was saved in this format
            if os.listdir(sim):
                pass
            continue
        hostname = sim
        hostname = hostname[hostname.find('/')+1:]
        hostname = hostname[:hostname.find('/')]
        run_base = os.path.join(newdir, hostname)
        run_num = 0
        while True:
            try:
                run_dir = run_base + '.' + str(run_num)
                os.makedirs(run_dir, mode=0o755)
                break
            except OSError:
                run_num = run_num + 1
        shutil.copytree(os.path.join(sim, 'input'), os.path.join(run_dir, 'input'))
        cpfile = lambda f: shutil.copyfile(os.path.join(sim, f),
                                           os.path.join(run_dir, f))
        try:
            cpfile('wlcsim.exe')
        except FileNotFoundError:
            print('Did not find wlcsim.exe in ' + sim)
        try:
            cpfile('run_parameter.pl')
        except FileNotFoundError:
            print('Did not find run_parameter.pl in ' + sim)
        try:
            cpfile('runsim.sh')
        except FileNotFoundError:
            print('Did not find runsim.sh in ' + sim)
        for i,finished_sims in enumerate(glob.glob(sim + '/savedata/*')):
            finished_dir = run_base + '.' + str(run_num + i + 1)
            shutil.copytree(run_dir, finished_dir)
            shutil.copytree(finished_sims, os.path.join(finished_dir, 'data'))
        shutil.copytree(os.path.join(sim, 'data'), os.path.join(run_dir, 'data'))




"""
Various routines that are too slow to run in real time are tabulated here.

Especially of note are Lena's ssWLC parameters code, which nobody has access to
anymore to the best of my knowledge (except Andy, who doesn't know how to use
it) and Tom's Rouse velocity cross-correlation code, which required an entire
rotation student to even run back when it did exist. I think it's saved under
some folder called ParConV4, but I haven't put in the time to try to figure out
if it's usable.
"""
import numpy as np
import pandas as pd
import scipy

from pathlib import Path

# precomupted velocity cross-correlation for rouse polymer
# from Lampo et al, was pre-computed on a grid...
deltas_ = np.linspace(-3, 3, 25)  # for corresponding values in logspace
alphas_ = np.linspace(0.25, 1, 31)  # steps of 0.025
tOverDeltas_ = np.linspace(0, 5, 501)  # steps of 0.01
vvcf_table_ = np.reshape(np.loadtxt(
        Path(__file__).parent / Path('vvcf_table.csv'), delimiter=','
    ), (31, 25, 501))


def calc_vel_corr_fixed_(tOverDelta, deltaOverTDeltaN, alpha):
    """Perform interpolation in logspace for "delta"/deltaOverTDeltaN."""
    deltaOverTDeltaN = np.log10(deltaOverTDeltaN)
    return scipy.interpolate.interpn((alphas_, deltas_, tOverDeltas_),
                                     vvcf_table_,
                                     (alpha, deltaOverTDeltaN, tOverDelta))


calc_vel_corr_fixed_.vvcf = None
frac_vel_corr = np.vectorize(calc_vel_corr_fixed_)

dsswlc_params = pd.read_csv(Path(__file__).parent / Path('dssWLCparams'),
                            sep=' ', header=None)
dsswlc_params.columns = ['DEL', 'EB', 'GAM', 'EPAR', 'EPERP', 'ETA', 'XIU',
                         'UNKNOWN']


def dsswlc_from_del(delta):
    """
    Get dsswlc sim params by interpolating Lena's tabulated values.

    Parameters
    ----------
    DEL : float
        Number of persistence lengths between each pair of beads.

    Returns
    -------
    ('EB', 'GAM', 'EPAR', 'EPERP', 'ETA', 'XIU', 'UNKNOWN')
        Parameters of the dsswlc for simulation.
    """
    return [np.interp(delta, dsswlc_params['DEL'], dsswlc_params[col])
            for col in dsswlc_params.columns[1:]]



"""
Some spare stochastic integrators.

Not currently used because if you pass a function as an argument you can't
@numba.jit that.

For rouse chains, Bruno (that's me) has tested the srk1 integrator and found it
to work extremely well (see notes in wlcsim.bd.rouse on suggested dt).

For WLC/ssWLC, Lena found it useful to use a scheme that was higher
deterministic order in time, to resolve the high elastic energies involved.
Bruno has not yet tested this explicitly.

You can find Lena's algorithm (hold brownian force constant over an rk4 step)
below. It is unlikely to be strongly convergent except if you subsample below
the actual desired time resolution.
"""
from numba import jit
import numpy as np


def rk4_thermal_lena(f, D, t, x0):
    """x'(t) = f(x(t), t) + Xi(t), where Xi is thermal, diffusivity D.

    x0 is x(t[0]).

    :math:`f: R^n x R -> R^n`
    """
    t = np.array(t)
    x0 = np.array(x0)
    x = np.zeros(t.shape + x0.shape)
    dts = np.diff(t)
    x[0] = x0
    dxdt = np.zeros((4,) + x0.shape)  # one for each RK step
    # at each step i, we use data (x,t)[i-1] to create (x,t)[i]
    # in order to make it easy to pull into a new functin later, we'll call
    # t[i-1] "t0", old x (x[i-1]) "x0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = dts[i-1]
        t0 = t[i-1]
        x0 = x[i-1]
        x_est = x0
        Fbrown = np.sqrt(2*D/(t[i]-t[i-1]))*np.random.normal(size=x0.shape)
        dxdt[0] = f(x0, t0) + Fbrown  # slope at beginning of time step
        x_est = x0 + dxdt[0]*(h/2)  # first estimate at midpoint
        dxdt[1] = f(x_est, t0 + (h/2)) + Fbrown  # estimated slope at midpoint
        x_est = x0 + dxdt[1]*(h/2)  # second estimate at midpoint
        dxdt[2] = f(x_est, t0 + (h/2)) + Fbrown  # second slope at midpoint
        x_est = x0 + dxdt[2]*h  # first estimate at next time point
        dxdt[3] = f(x_est, t0 + h) + Fbrown  # slope at end of time step
        # final estimate is weighted average of slope estimates
        x[i] = x0 + h*(dxdt[0] + 2*dxdt[1] + 2*dxdt[2] + dxdt[3])/6
    return x


def rk4_thermal_bruno(f, D, t, x0):
    r"""
    Test new method: Attempt to keep :math:`\omega` constant.

    WARNING: does not converge strongly (autocorrelation function seems
    higher than should be for OU process...), as is...x'(t) = f(x(t), t) +
    Xi(t), where Xi is thermal, diffusivity D

    x0 is x(t[0]).

    :math:`f: R^n x R -> R^n`
    """
    t = np.array(t)
    x0 = np.array(x0)
    xsize = x0.shape
    x = np.zeros(t.shape + x0.shape)
    dts = np.diff(t)
    x[0] = x0
    dxdt = np.zeros((4,) + x0.shape)  # one for each RK step
    Fbrown = np.sqrt(2*D / ((t[1] - t[0])/2))
    # at each step i, we use data (x,t)[i-1] to create (x,t)[i]
    # in order to make it easy to pull into a new functin later, we'll call
    # t[i-1] "t0", old x (x[i-1]) "x0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = dts[i-1]
        t0 = t[i-1]
        x0 = x[i-1]
        x_est = x0
        dxdt[0] = f(x0, t0) + Fbrown  # slope at beginning of time step
        # random force estimate at midpoint
        Fbrown = np.sqrt(2*D / ((t[i]-t[i-1])/2))*np.random.normal(size=xsize)
        x_est = x0 + dxdt[0]*(h/2)  # first estimate at midpoint
        dxdt[1] = f(x_est, t0 + (h/2)) + Fbrown  # estimated slope at midpoint
        x_est = x0 + dxdt[1]*(h/2)  # second estimate at midpoint
        dxdt[2] = f(x_est, t0 + (h/2)) + Fbrown  # second slope at midpoint
        x_est = x0 + dxdt[2]*h  # first estimate at next time point
        # random force estimate at endpoint (and next start point)
        Fbrown = np.sqrt(2*D / ((t[i]-t[i-1])/2))*np.random.normal(size=xsize)
        dxdt[3] = f(x_est, t0 + h) + Fbrown  # slope at end of time step
        # final estimate is weighted average of slope estimates
        x[i] = x0 + h*(dxdt[0] + 2*dxdt[1] + 2*dxdt[2] + dxdt[3])/6
    return x


def euler_maruyama(f, D, t, x0):
    t = np.array(t)
    x0 = np.array(x0)
    x = np.zeros(t.shape + x0.shape)
    dts = np.diff(t)
    x[0] = x0
    # at each step i, we use data (x,t)[i-1] to create (x,t)[i]
    # in order to make it easy to pull into a new functin later, we'll call
    # t[i-1] "t0", old x (x[i-1]) "x0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = dts[i-1]
        t0 = t[i-1]
        x0 = x[i-1]
        Fbrown = np.sqrt(2*D/(t[i]-t[i-1]))*np.random.normal(size=x0.shape)
        x[i] = x0 + h*(Fbrown + f(x0, t0))
    return x


def srk1_roberts(f, D, t, x0):
    r"""From wiki, from A. J. Roberts. Modify the improved Euler scheme to
    integrate stochastic differential equations. [1], Oct 2012.

    If we have an Ito SDE given by

    .. math::

        d\vec{X} = \vec{a}(t, \vec{X}) + \vec{b}(t, \vec{X}) dW

    then

    .. math::

        \vec{K}_1 = h \vec{a}(t_k, \vec{X}_k) + (\Delta W_k - S_k\sqrt{h}) \vec{b}(t_k, \vec{X}_k)
        \vec{K}_2 = h \vec{a}(t_{k+1}, \vec{X}_k + \vec{K}_1) + (\Delta W_k - S_k\sqrt{h}) \vec{b}(t_{k+1}, \vec{X}_k + \vec{K}_1)
        \vec{X}_{k+1} = \vec{X}_k + \frac{1}{2}(\vec{K}_1 + \vec{K}_2)

    where :math:`\Delta W_k = \sqrt{h} Z_k` for a normal random :math:`Z_k \sim
    N(0,1)`, and :math:`S_k=\pm1`, with the sign chosen uniformly at random
    each time."""
    t = np.array(t)
    x0 = np.array(x0)
    x = np.zeros(t.shape + x0.shape)
    dts = np.diff(t)
    # -1 or 1, p=1/2
    S = 2*(np.random.random_sample(size=t.shape) < 0.5) - 1
    x[0] = x0
    # at each step i, we use data (x,t)[i-1] to create (x,t)[i]
    # in order to make it easy to pull into a new functin later, we'll call
    # t[i-1] "t0", old x (x[i-1]) "x0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = dts[i-1]
        t0 = t[i-1]
        x0 = x[i-1]
        dW = np.random.normal(size=x0.shape)
# D = sigma^2/2 ==> sigma = np.sqrt(2*D)
        Fbrown = np.sqrt(2*D/h)*(dW - S[i])
        # estimate for slope at interval start
        K1 = f(x0, t0) + Fbrown
        Fbrown = np.sqrt(2*D/h)*(dW + S[i])
        # estimate for slope at interval end
        K2 = f(x0+h*K1, t0+h) + Fbrown
        x[i] = x0 + h * (K1 + K2)/2
    return x


# simple test case
def ou(x0, t, k_over_xi, D, method=rk4_thermal_lena):
    "simulate ornstein uhlenbeck process with theta=k_over_xi and sigma^2/2=D"
    def f(x,t):
        return -k_over_xi*x
    return method(f, D=D, t=t, x0=x0)


@jit(nopython=True)
def _get_scalar_corr(X):
    "fast correlation calculation for testing"
    num_samples, num_t = X.shape
    corr = np.zeros((num_t,))
    count = np.zeros((num_t,))
    for i in range(num_samples):
        for j in range(num_t):
            for k in range(j, num_t):
                corr[k-j] = corr[k-j] + X[i,k]*X[i,j]
                count[k-j] = count[k-j] + 1
    return corr, count


@jit(nopython=True)
def _get_vector_corr(X):
    "fast correlation calculation for testing"
    num_samples, num_t, d = X.shape
    corr = np.zeros((num_t,))
    count = np.zeros((num_t,))
    for i in range(num_samples):
        for j in range(num_t):
            for k in range(j, num_t):
                corr[k-j] = corr[k-j] + X[i,k]@X[i,j]
                count[k-j] = count[k-j] + 1
    return corr, count


@jit(nopython=True)
def _get_bead_msd(X, k=None):
    """center bead by default

    for 1e4-long time arrays, this takes ~10-30s on my laptop"""
    num_t, num_beads, d = X.shape
    if k is None:
        k = max(0, num_beads/2 - 1)
    k = int(k)
    ta_msd = np.zeros((num_t,))
    count = np.zeros((num_t,))
    for i in range(num_t):
        for j in range(i, num_t):
            ta_msd[j-i] += (X[j,k] - X[i,k])@(X[j,k] - X[i,k])
            count[j-i] += 1
    return ta_msd, count


@jit(nopython=True)
def _msd(x):
    result = np.zeros_like(x)
    for delta in range(1,len(x)):
        for i in range(delta,len(x)):
            result[delta] += (x[i] - x[i-delta])**2
        result[delta] = result[delta] / (len(x) - delta)
    return result


# test different integrators below on simply OU process
def test_ou_autocorr(method=srk1_roberts):
    import scipy.stats
    import matplotlib.pyplot as plt
    k = 2
    xi = 4
    kbT = 1
    D = kbT/xi
    k_over_xi = k/xi
    x0 = scipy.stats.norm.rvs(scale=np.sqrt(D/k_over_xi), size=(10_000,))
    t = np.linspace(0, 1e2, 1e3+1)
    X = ou(x0, t, k_over_xi, D, method=method)
    assert(np.abs(np.var(X) - D/k_over_xi)/(D/k_over_xi) < 0.1)
    corr, count = _get_scalar_corr(X.T)
    err = corr/count - (kbT/k)*np.exp(-k_over_xi*t)
    plt.figure()
    plt.plot(t, err)
    plt.figure()
    plt.hist(X[-100:].flatten(), bins=100, density=1)
    x = np.linspace(-3, 3, 100)
    plt.plot(x, scipy.stats.norm(scale=np.sqrt(D/k_over_xi)).pdf(x))
    return X



""" This module "understands" wlcsim.exe's output, and will make it into
nice, pretty numpy/pandas arrays for you."""
import numpy as np
import pandas as pd
import os
import logging
import itertools
import glob
import re
from .input import ParsedInput
from .utils.cached_property import cached_property
import shutil

from numba import double
from numba.decorators import jit

logger = logging.getLogger(__name__)

# keep these arround for cache reasons
# initialize with size 1
i_arr = np.array([[1]])
j_arr = np.array([[1]])
ui = (np.array([0]), np.array([0]))

@jit
def pairwise_distances3(r):
    """
    Returns np.array of pairwise distances of points in R^3 stored in a
    3-by-num_points numpy array.
    """
    n = r.shape[1]
    dr = np.empty((n, n), dtype=np.float)
    for i in range(n):
        for j in range(n):
            d = 0.0
            for k in range(3):
                tmp = r[k, i] - r[k, j]
                d += tmp*tmp
            dr[i, j] = np.sqrt(d)
    return dr


def df_from_coltimes_method3(coltimes):
    # for caching
    global i_arr
    global j_arr
    global ui
    # method 3: manipulate numpy arrays directly. about same speed as
    # method #2, but insanely more complex to reason about
    N = coltimes.shape[0]
    if not np.all(j_arr.shape == coltimes.shape):
        j_arr = np.linspace(0, N-1, N)
        j_arr = np.tile(j_arr, (N, 1))
        i_arr = j_arr.T
        j_arr = np.reshape(j_arr, (np.product(j_arr.shape), 1))
        i_arr = np.reshape(i_arr, (np.product(i_arr.shape), 1))
        ui = np.triu_indices(N, k=0)
    coltimes[ui] = np.nan
    coltimes_lin = np.reshape(coltimes, (np.product(coltimes.shape), 1))
    to_keep_idx = np.logical_not(np.isnan(coltimes_lin))
    coltimes_lin = coltimes_lin[to_keep_idx]
    i_lin = i_arr[to_keep_idx]
    j_lin = j_arr[to_keep_idx]
    coltimes = np.vstack((coltimes_lin, i_lin, j_lin)).T
    if coltimes.ndim == 1:
        coltimes = coltimes.reshape(1, coltimes.shape[0])
    return pd.DataFrame(coltimes, columns=('coltime', 'i', 'j'))

def df_from_coltimes_method2(coltimes):
    ## # turns out this is slow compared to pandas or textreader
    ##coltimes = np.loadtxt(coltimes_file)
    # method 2: still slow from creating Series?
    coltimes_serieses = []
    # make a Series out of the lower triangular part of coltimes matrix
    for i in range(coltimes.shape[0]):
        for j in range(i-1):
            # method 2: still slow from creating Series?
            coltimes_serieses.append(pd.Series({
                    'i': i, 'j': j,
                    'coltime': coltimes[i][j]}))
            # # method 1: this causes N reallocations, super slow
            # self.coltimes = self.coltimes.append(
            #     pd.Series({'i': i, 'j': j,
            #                'coltime': coltimes[i][j]}),
            #     ignore_index=True
            # )
    # method 2: still slow from creating Series?
    return pd.DataFrame(coltimes_serieses)

def df_from_coltimes_method1(coltimes):
    df = pd.DataFrame()
    # make a Series out of the lower triangular part of coltimes matrix
    for i in range(coltimes.shape[0]):
        for j in range(i-1):
            # method 1: this causes N reallocations, super slow
            self.coltimes = df.append(
                pd.Series({'i': i, 'j': j,
                           'coltime': coltimes[i][j]}),
                ignore_index=True
            )
    # method 2: still slow from creating Series?
    return df

def non_overlapping_pairs(length, spacing, max_pairs=None, min_pairs=0):
    """Return the "most spaced" possible sets of pairs of indices that are a
    specific spacing apart, for getting statistics that aren't biased by the
    overlap between neighboring segments of the polymer."""
    # use the number of possible left endpoints to get the num possible pairs
    # num_possible_pairs = np.floor((length - spacing)/spacing).astype(int)
    # jk, number possible pairs is easy
    num_possible_pairs = np.floor(length/spacing).astype(int)
    if num_possible_pairs < min_pairs:
        raise ValueError("Requested too large of a spacing for a given minimum "
                         + "number of overlapping pairs")
    # by default, return all possible non-overlapping pairs
    if max_pairs is None or max_pairs > num_possible_pairs:
        max_pairs = num_possible_pairs
    num_leftover_indices = length - max_pairs*spacing
    # first find skip between steps, rounded nicely, we'll use leftovers from
    # rounding to pad the start and end
    # num skips is of course num_pairs-1, fenceposts problem
    if max_pairs == 1:
        skip_per_step = 0 # only one valid pair, so no skip between them
    else:
        skip_per_step = int(num_leftover_indices/(max_pairs-1))
    num_leftover_indices -= skip_per_step*(max_pairs-1)
    # the leftover indices should pad the outside of the things, so we just use
    # half of them to pad on the left
    num_leftover_indices = int(num_leftover_indices/2)
    left, right = (0, spacing)
    left, right = (left + num_leftover_indices, right + num_leftover_indices)
    yield (left, right)
    for i in range(max_pairs-1):
        # if num_leftover_indices > 0:
        #     num_leftover_indices -= 1
        #     left = right+skip_per_step+1
        #     right = left + spacing
        left, right = (right + skip_per_step, right + skip_per_step + spacing)
        yield (left, right)

class Sim:
    """
    Sim provides an interface to the heterogenous data produced by a
    simulation that saves coltimes, positions, and momentums.

    Unless specified during initialization, the coltimes, r, and u are loaded
    on demand upon first access.

    Accessing properties of this class for the first time can return an IOError
    if a file is moved, or doesn't exist. Afterwards, the values will be cached.

    args:
        input_file : str - relative or absolute location of input file
        file_suffix : str - custom suffix on saved filenames, e.g.  corresponding to the particular replica that should be loaded
    """

    def __init__(self, sim_path, input_file=None, file_suffix=None,
                 overwrite_coltimes=False, load_method=3, rfile_base='r'):
        # since these are all lazily evaluated, we throw exceptions right away
        # if the files we need do not exist, to prevent long running programs
        # from throwing an exception midway
        if not os.path.isdir(sim_path):
            raise FileNotFoundError('Directory (' + sim_path +
                                    ') provided when constructing'
                                    ' wlcsim.data.Sim does not exist!')
        self.sim_path = sim_path
        self.sim_dir = os.path.basename(sim_path)
        self.input_file = Sim.guess_input_name(sim_path, input_file)
        # without input file, we don't really have a chance of using the
        # simulation data from a simulation anyway, so we can safely error out
        # here
        if not os.path.isfile(self.input_file):
            raise FileNotFoundError(self.input_file)
        self.data_dir = os.path.join(sim_path, 'data')
        self.coltimes_file = os.path.join(self.data_dir, 'coltimes')
        self.overwrite_coltimes = overwrite_coltimes
        self._rfile_base = os.path.join(self.data_dir, rfile_base)
        self._ufile_base = os.path.join(self.data_dir, 'u')
        self._file_suffix = 'v0' if file_suffix is None else str(file_suffix)
        self.load_method = load_method
        self._has_loaded_r = False

    def guess_input_name(sim_path, input_file):
        bruno_input_name = os.path.join(sim_path, 'input', 'input')
        quinn_input_name = os.path.join(sim_path, 'input', 'params')
        def_input_name = os.path.join(sim_path, 'src', 'defines.inc')
        if input_file:
            if os.path.isfile(input_file):
                return input_file
            elif os.path.is_file(os.path.join(sim_path, input_file)):
                return os.path.join(sim_path, input_file)
            elif os.path.is_file(os.path.join(sim_path, 'input', input_file)):
                return os.path.join(sim_path, 'input', input_file)
        if os.path.isfile(def_input_name):
            return def_input_name
        elif os.path.isfile(bruno_input_name):
            return bruno_input_name
        elif os.path.isfile(quinn_input_name):
            return quinn_input_name
        else:
            return None

    @cached_property
    def parsed_input(self):
        return ParsedInput(self.input_file)

    @cached_property
    def coltimes(self):
        coltimes = pd.read_csv(self.coltimes_file, delim_whitespace=True,
                                header=None).values
        coltimes = self.df_from_coltimes(coltimes, method=self.load_method)
        return coltimes

    @cached_property
    def r0(self):
        """ Numpy array of positions of each particle pre-first time point. """
        rfile_name = self._rfile_base + str(0) + self._file_suffix
        r0 = pd.read_csv(rfile_name, delim_whitespace=True, header=None)
        return r0.values.reshape((3, self.num_polymers, self.num_beads))

    @cached_property
    def rend(self):
        """ Numpy array of positions of each particle pre-first time point. """
        last_saved_ti = np.max(self.rtime_idxs)
        rfile_name = self._rfile_base + str(last_saved_ti) + self._file_suffix
        rend = pd.read_csv(rfile_name, delim_whitespace=True, header=None)
        return rend

    @property
    def rend2end(self):
        # there's a fast way to do this if there's only one polymer
        #TODO use known linesize to extend this to work no matter how many
        # polymers there are
        if self.num_polymers == 1:
            rend2end = np.empty((self.num_time_points, 3, self.num_polymers))
            rend2end[:] = np.nan
            for ti in self.rtime_idxs:
                rfile_name = self._rfile_base + str(ti) + self._file_suffix
                with open(rfile_name, 'rb') as f:
                    first_line = f.readline().rstrip()
                    line_len = len(first_line)
                    r1 = np.array([float(rij) for rij in first_line.split()])
                    f.seek(-line_len*2, os.SEEK_END)
                    last_line = f.readlines()[-1].rstrip()
                    rN = np.array([float(rij) for rij in last_line.split()])
                    rend2end[ti,:,0] = rN - r1
        else:
            rend2end = self.r[:,:,:,-1] - self.r[:,:,:,0]
        return rend2end

    @cached_property
    def u0(self):
        """ Numpy array of positions of each particle at each time point. """
        ufile_name = self._ufile_base + str(0) + self._file_suffix
        u0 = pd.read_csv(rfile_name, delim_whitespace=True, header=None)
        return u0.values.reshape((3, self.num_polymers, self.num_beads))

    @cached_property
    def r(self):
        """ Numpy array of positions of each particle at each time point.
        shape:
        r.shape == (self.num_time_points, 3, self.num_polymers, self.num_beads)
        """
        r = np.full((self.num_time_points, 3, self.num_polymers,
                     self.num_beads), np.nan)
        for i in self.rtime_idxs:
            rfile_name = self._rfile_base + str(i) + self._file_suffix
            ri = pd.read_csv(rfile_name, delim_whitespace=True, header=None)
            r[i,:,:,:] = ri.values[:,0:3].T.reshape((1, 3, self.num_polymers, self.num_beads))
        self._has_loaded_r = True
        return r

    @cached_property
    def u(self):
        """ Numpy array of positions of each particle at each time point. """
        u = np.full((self.num_time_points, 3, self.num_polymers,
                     self.num_beads), 0.0)
        for i in self.utime_idxs:
            ufile_name = self._ufile_base + str(i) + self._file_suffix
            ui = pd.read_csv(ufile_name, delim_whitespace=True, header=None)
            u[i,:,:,:] = ui.values.T.reshape((1, 3, self.num_polymers,
                                            self.num_beads))
        return u

    @property
    def num_time_points(self):
        return int(self.params['NUMSAVEPOINTS']) + 1 # account for time 0

    @property
    def num_polymers(self):
        try:
            return int(self.params['NP'])
        except:
            return 1

    @property
    def num_beads(self):
        return int(self.params['NB'])

    @cached_property
    def center_of_mass(self):
        # "equal mass" beads, so just average positions
        return np.mean(self.r, axis=3)

    @cached_property
    def rtime_idxs(self):
        rfiles = glob.glob(self._rfile_base + '[0-9]*' + self._file_suffix)
        return np.sort(self._time_idxs(self._rfile_base, rfiles))

    @cached_property
    def utime_idxs(self):
        ufiles = glob.glob(self._ufile_base + '[0-9]*' + self._file_suffix)
        return np.sort(self._time_idxs(self._ufile_base, ufiles))

    def _time_idxs(self, base, file_names):
        number_from_filename = re.compile(re.escape(base) + '([0-9]+)')
        return np.array([int(re.search(number_from_filename, file_name).groups()[0])
                         for file_name in file_names])

    @cached_property
    def param_names(self):
        param_names, _ = self.parsed_input
        return param_names

    @property # caching won't help a call that's just a redirection
    def params(self):
        return self.parsed_input.params

    @cached_property
    def rtimes(self):
        return self.rtime_idxs * self.params['DT']

    @cached_property
    def utimes(self):
        return self.utime_idxs * self.params['DT']

    @cached_property
    def polymer_length(self):
        """The sum of interbead distances in the polymer at each time"""
        sqd = np.sum(np.power(self.r[:,:,:,1:] - self.r[:,:,:,0:-1], 2),
                     axis=(1,3))
        return np.sqrt(sqd)


    def dr_mean_by_distance(self, lengths, max_samples_per_len=1000):
        distances = np.full((len(lengths), self.num_time_points, self.num_polymers), np.nan)
        for i,length in enumerate(lengths):
            mean_dist = 0
            num_dists = 0
            for left, right in non_overlapping_pairs(self.polymer_length,
                    length, max_pairs=max_samples_per_len):
                mean_dist += np.linalg.norm(self.r[:,:,:,right] - self.r[:,:,:,left], axis=1)
                num_dists += 1
            distances[i,:,:] = np.squeeze(mean_dist/num_dists)
        return distances

    # the properties of coltimes are all coded as a function of the linear
    # distance between the beads of interest, as opposed to as a function of
    # their absolute location, although that may not be super good?
    @property
    def fraction_collided(self):
        return 1.0 - self.fraction_uncollided

    @property
    def fraction_uncollided(self):
        return np.array([float(self.num_uncollided[i])/float(self.num_possible[i])
                         for i in range(length(self.num_possible))])

    @property
    def num_uncollided(self):
        c = self.coltimes
        distances = c['linear_distance'].unique()
        return np.array([sum(pd.isnull(c.loc[c['linear_distance'] == i, 'coltime']))
                         for i in distances])

    @property
    def num_collided(self):
        return self.num_possible - self.num_uncollided

    @property
    def num_possible(self):
        c = self.coltimes
        distances = c['linear_distance'].unique()
        return np.array([len(c.loc[c['linear_distance'] == i, 'coltime'])
                         for i in distances])

    def position_in_axis(self, axis=0):
        """Project position as a function of N along the chain into one of it's
        dimensions to see how it's oriented.

        Useful for e.g. visualizing if bacterial DNA in confinement aligns
        linearly along the cell's length as in e.g. Caulobacter."""
        return np.squeeze(self.r[:,axis,:,:])

    def df_from_coltimes(self, coltimes, method=3):
        # coltimes_file = os.path.join(self.data_dir, 'coltimes.csv')
        # if os.path.isfile(coltimes_file) and not self.overwrite_coltimes:
        #     coltimes = pd.read_csv(coltimes_file, index_col=0)
        #     return coltimes
        # elif method == 1:
        if method == 1:
            coltimes = df_from_coltimes_method1(coltimes)
        elif method == 2:
            coltimes = df_from_coltimes_method2(coltimes)
        elif method == 3:
            coltimes = df_from_coltimes_method3(coltimes)
        else:
            raise KeyError('load_method=' + str(method) + ' does not exist!')
        nan_fill_coltimes(coltimes)
        self.add_useful_cols_to_coltimes(coltimes)
        # coltimes.to_csv(coltimes_file)
        return coltimes

    def add_useful_cols_to_coltimes(self, coltimes):
        """Only works on coltimes that come from combine_coltimes_csvs."""
        dr = float(self.params['L'])/(float(self.params['NB']) - 1)
        coltimes['linear_distance'] = dr*np.abs(coltimes['i'] - coltimes['j'])

class ReplicaSim:
    """ReplicaSim(path).sims == [Sim(path, file_suffix=suffix)
                                 for suffix in path's data folder]
    """

    known_data_prefixes = ['r', 'u', 'phi']

    def __init__(self, sim_path, *args, **kwargs):
        if 'file_suffix' in kwargs:
            raise ValueError('Cannot pass ReplicaSim a specific file_suffix, just use Sim for this.')
        self.dummy_sim = Sim(sim_path, *args, **kwargs)
        self.file_suffixes = list(ReplicaSim.data_file_suffixes(self.dummy_sim.data_dir))
        if self.file_suffixes:
            self.sims = [Sim(sim_path, *args, file_suffix=suffix, **kwargs)
                        for suffix in self.file_suffixes]
        else:
            self.sims = [self.dummy_sim]

    @staticmethod
    def data_file_suffixes(data_dir):
        """Get file suffixes in given data_dir by looking for files that look
        like standard data files, only looking at time points 0 and 1 for
        speed."""
        known_suffixes = set()
        for base in ReplicaSim.known_data_prefixes:
            file_name0 = glob.glob(os.path.join(data_dir, base) + '0*')
            file_name1 = glob.glob(os.path.join(data_dir, base) + '1*')
            file_names = itertools.chain(file_name0, file_name1)
            suffix_from_filename = re.compile(re.escape(base) + '[01][0-9]*(.*)')
            for file_name in file_names:
                result = re.search(suffix_from_filename, file_name)
                if result is None:
                    raise LookupError('Filename ' + file_name + ' should be ' \
                                        + 'a save file with base ' + base \
                                        + ' but that regex failed.')
                suffix = result.groups()[0]
                if suffix in known_suffixes:
                    continue
                known_suffixes.add(suffix)
                yield suffix

class Scan:
    """Can tell you what dirs in a run_dir are simulation directories, and what
    all the input files said, and more. """

    def __init__(self, run_dir, limit=None, complete=False, *args, **kwargs):
        self.limit = limit
        self.complete = complete
        self.path = run_dir
        self._prepend_path = lambda d: os.path.join(self.path, d)
        self._is_simdir = lambda d: Scan.is_simdir(d, self.path)
        self._is_complete = lambda d: Scan.is_complete(d, self.path)
        self._sim_args = args
        self._sim_kwargs = kwargs
        # default to behavior that's not default for when you are only looking
        # at an individual simulation, and specify the file save suffix
        # explicitly
        if not kwargs or 'file_suffix' not in kwargs:
            self.infer_file_suffix = True
        else:
            self.infer_file_suffix = False

    @property
    def sim_dirs(self):
        if self.complete:
            is_sim = lambda d: self._is_simdir(d) and self._is_complete(d)
        else:
            is_sim = self._is_simdir
        sim_dirs = filter(is_sim, os.listdir(self.path))
        if self.limit is None:
            return list(sim_dirs)
        else:
            return list(itertools.islice(sim_dirs, self.limit))

    @property
    def sim_paths(self):
        return list(map(self._prepend_path, self.sim_dirs))

    @staticmethod
    def is_simdir(dirname, run_dir):
        sim_name_re = re.compile('.*\.[0-9]+')
        abs_dirname = os.path.join(run_dir, dirname)
        input_file = os.path.join(abs_dirname, 'input', 'input')
        return sim_name_re.search(dirname) \
                and os.path.isdir(abs_dirname) \
                and os.path.isfile(input_file)

    @staticmethod
    def is_complete(dirname, run_dir):
        abs_dirname = os.path.join(run_dir, dirname)
        complete_file = os.path.join(abs_dirname, 'complete')
        return os.path.isfile(complete_file)

    # @cached_property
    # a generator cannot be a cached property, else it can only be used once
    @property
    def sims(self):
        for folder in self.sim_paths:
            if self.infer_file_suffix:
                for sim in ReplicaSim(folder, *self._sim_args, **self._sim_kwargs).sims:
                    yield sim
            else:
                yield Sim(folder, *self._sim_args, **self._sim_kwargs)

    @cached_property
    def inputs(self):
        df = pd.DataFrame([sim.params for sim in self.sims])
        df.index = [sim.sim_dir for sim in self.sims]
        return df

    @cached_property
    def all_param_values(self):
        return {col: self.inputs[col].unique() for col in self.inputs.columns}

    @property
    def run_dir(self):
        return self.path

    @property
    def scan_dir(self):
        return self.path

    # from back when we used this class to combine coltimes's from a scan
    # into one big pandas df
    # def calculate_linear_distance(self):
        # # actual distance along polymer is dx*(number_separating)
        # self.coltimes['dx'] = self.coltimes['L']/(self.coltimes['NB'] - 1)
        # self.coltimes['linear_distance'] = self.coltimes['dx']*np.abs(self.coltimes['i'] - self.coltimes['j'])
    # def calculate_for_each_param(self, f):
        # """unimplemented"""
        # pass

    @property
    def rend2end(self):
        """End bead distances for each simulation as numpy array"""
        example_sim = self.sims[0]
        rend2end = np.empty([len(self.sims), example_sim.num_time_points,
                             3, example_sim.num_polymers])
        rend2end[:] = np.nan
        print('This is a time consuming function, I will print out a number'
              'for each sim I process.')
        for i,sim in enumerate(self.sims):
            print(i)
            rend2end[i,:,:,:] = sim.rend2end
        return rend2end

    def write_coltimes_csvs(self, overwrite=False, print_progress=False, *args, **kwargs):
        for folder in self.sim_paths:
            if print_progress:
                print(folder + ' ... ', end='')
            coltimes_file = os.path.join(folder, 'coltimes.csv')
            if not overwrite and os.path.isfile(coltimes_file):
                continue
            try:
                sim = Sim(folder, *args, **kwargs)
            except FileNotFoundError:
                logger.warning('write_coltimes_csvs: Unable to construct sim from' + folder)
                if print_progress:
                    print('FAILED!')
                continue
            try:
                sim.coltimes.to_csv(coltimes_file)
            except OSError:
                logger.warning('No coltimes file in: ' + folder)
                if print_progress:
                    print('FAILED!')
                continue
            if print_progress:
                print('complete!')

def write_initial_dists_csvs(run_dir, overwrite=True, *args, **kwargs):
    for folder in Scan(run_dir).sim_paths:
        dists_file = os.path.join(folder, 'initial_dists.csv')
        if not overwrite and os.path.isfile(dists_file):
            continue
        try:
            sim = Sim(folder, *args, **kwargs)
        except FileNotFoundError:
            logger.warning('write_initial_dists_csvs: Unable to construct sim from' + folder)
            continue
        try:
            np.savetxt(dists_file, pairwise_distances3(sim.r0), delimiter=',')
        except OSError:
            logger.warning('r0 not found for: ' + folder)
            continue
        except ValueError:
            logger.warning('r0 file is empty in: ' + folder)
            continue

def write_coltimes_pkls(run_dir, *args, **kwargs):
    for folder in Scan(run_dir).sim_paths:
        try:
            sim = Sim(folder, *args, **kwargs)
        except FileNotFoundError:
            logger.warning('write_coltimes_pkls: Unable to construct sim from' + folder)
        try:
            sim.coltimes.to_pickle(os.path.join(folder, 'coltimes.pkl'))
        except OSError:
            logger.warning('No coltimes file in: ' + folder)
            continue
        except ValueError:
            logger.warning('coltimes file is empty in: ' + folder)
            continue

# def combine_coltimes_pkls(run_dir, *args, **kwargs):
#     pkl_glob = os.path.join(run_dir, '*', 'coltimes.pkl')
#     for coltimes_pkl in glob.glob(pkl_glob):

def combine_coltimes_csvs(run_dir, conditionals=[],
                          outfile_name='coltimes.csv'):
    """takes a directory that scan_wlcsim.py has output to and an optional list
    of functions that takes a coltimes array and decides whether it may be
    used, and an output location"""
    #TODO: finish converting to following:
    # csv_files = [os.path.join(folder, 'coltimes.csv')
    #              for folder in Scan(run_dir).sim_paths]
    # csv_files = [f for f in csv_files if os.path.isfile(f)]
    csv_glob = os.path.join(run_dir, '*', 'coltimes.csv')
    csv_files = iter(glob.glob(csv_glob))
    first_file = next(csv_files)
    coltimes = pd.read_csv(first_file, index_col=0)
    coltimes['run_name'] = os.path.basename(os.path.dirname(first_file))
    # loop and a half. create file uncoditinoally on pre-first iteration so
    # that we can append from then on
    input_file = os.path.join(os.path.dirname(first_file), 'input', 'input')
    parsed_input = ParsedInput(input_file)
    for param in parsed_input.ordered_param_names:
        coltimes[param] = parsed_input.params[param]
    coltimes.to_csv(os.path.join(run_dir, 'coltimes.csv'), index=False)
    # now append the remaining
    for csv_file in csv_files:
        print(csv_file)
        coltimes = pd.read_csv(csv_file, index_col=0)
        coltimes['run_name'] = os.path.basename(os.path.dirname(csv_file))
        # throw away all coltimes that don't pass our conditional checks
        keep_this_coltimes = True
        for conditional in conditionals:
            if not conditional(coltimes):
                keep_this_coltimes = False
                break
        if not keep_this_coltimes:
            continue
        input_file = os.path.join(os.path.dirname(first_file), 'input', 'input')
        parsed_input = ParsedInput(input_file)
        for param in parsed_input.ordered_param_names:
            coltimes[param] = parsed_input.params[param]
        coltimes.to_csv(os.path.join(run_dir, outfile_name),
                                     index=False, mode='a', header=False)


def nan_fill_coltimes(coltimes):
    """Takes a raw coltimes pandas array from e.g. write_coltimes_csvs and puts
    nan's where the -1.0's are, i.e. where the simulation did not go out far
    enough."""
    coltimes.loc[coltimes['coltime'] == -1.0, 'coltime'] = np.nan

def non_overlapping_coltimes(cdf):
    """Takes a coltimes dataframe from write_coltimes_csvs/combine_coltimes_csvs
    and retains only non-overlapping coltimes examples."""
    L = cdf.L.unique()
    if len(L) > 1:
        raise ValueError('Not coded to deal with simulations of different length polymers.')
    L = L[0]
    # build final index by accumulating the indices that we should keep
    # make a convenience column to use isin, make smaller index first
    cij = pd.Series(list(zip(cdf.j, cdf.i)))
    # start with something that's all false of the right size
    non_overlapping_ix = cdf.linear_distance < 0
    for ld in cdf.linear_distance.unique():
        valid_indices = set(non_overlapping_pairs(L, ld))
        non_overlapping_ix |= cij.isin(valid_indices)
    return cdf.loc[non_overlapping_ix]


def combine_complete_coltime_csvs(run_dir):
    combine_coltimes_csvs(run_dir, [], 'complete_coltimes.csv')

def add_useful_cols_to_coltimes(coltimes):
    """Only works on coltimes that come from combine_coltimes_csvs."""
    coltimes['deltaR'] = coltimes['L']/(coltimes['NB'] - 1)
    coltimes['linear_distance'] = coltimes['deltaR']*np.abs(coltimes['i'] - coltimes['j'])

def combine_initial_dists_csvs(run_dir, *args, **kwargs):
    """Not yet implemented."""
    pass

def aggregate_r0_group_NP_L0(run_dir, overwrite=True, *args, **kwargs):
    e2e_folder = os.path.join(run_dir, 'data', 'end_to_end_r0')
    os.makedirs(e2e_folder, mode=0o755, exist_ok=True)
    existing_r0_files = {}
    for folder in glob.glob(os.path.join(run_dir, '*.*')):
        if not os.path.isdir(folder):
            continue
        try:
            sim = Sim(folder, *args, **kwargs)
        except FileNotFoundError:
            logger.warning('aggregate_r0_group_NP_L0: Unable to construct sim from ' + folder)
            continue
        # #beads
        N = sim.params['NB']
        # length in units of simulation
        L = sim.params['L']
        r0_agg_file = os.path.join(e2e_folder, 'r0_' + str(N) + '_' + str(L) + '.csv')
        r0_file = os.path.join(folder, 'data', 'r0')
        if os.path.isfile(r0_agg_file) and not overwrite:
            continue
        if r0_agg_file not in existing_r0_files:
            existing_r0_files[r0_agg_file] = True
            try:
                shutil.copyfile(r0_file, r0_agg_file)
            except OSError:
                logger.warning('r0 not found for: ' + folder)
                open(r0_agg_file, 'a').close() # make sure file exists later in loop
            continue
        else:
            try:
                r0_in = open(r0_file, 'r')
            except:
                logger.warning('r0 not found for: ' + folder)
                continue
            with open(r0_agg_file, 'a') as r0_out:
                for line in r0_in:
                    r0_out.write(line)
            r0_in.close()



"""
BD of two homologously-linked linear Rouse polymers.

See `homolog_points_to_loops_list` for a description of how the homologous
network structure is handled.

A specialized
"""
from ..plot import PolymerViewer
from .rouse import recommended_dt
from .init_beads import init_homolog_rouse_conf
from .forces import f_conf_ellipse, f_elas_homolog_rouse, f_tether_ellipse

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
from numba import jit


# The constraints for dt here are the same as any other Rouse polymer.
recommended_dt = recommended_dt


def points_to_loops_list(N, homolog_points):
    r"""Create loops list for jit_rouse_linked.

    Parameters
    ----------
    N : int
        number of beads in each polymer if they were unbound
    homolog_points : (L,) array_like
        list of loci indices :math:`{l_i}_1^L, l_i\in[1,N]` (beads) that are
        homologously paired.

    Returns
    -------
    N_tot : int
        final size of the x0 array for the simulation.
    loop_list : (L,4) np.array
        data structure used by jit_rouse_linked to correctly compute the spring
        forces in the network.

    Notes
    -----
    You can use the output like

    >>> k1l, k1r, k2l, k2r = loop_list[i]

    where "k[i][l,r]" refers to the bead demarcating the [l]eft or [r]ight end
    of a given "homologous loop" on each of the polymers polymer
    :math:`i\in[1,2]`.

    Namely, the connectivity structure of our polymer "network" is that for
    each `k1l, k1r, k2l, k2r in loop_list`, the beads `k1l` thru `k1r` are
    linked sequentially (as are `k2l` thru `k2r`), but the beads `k1[l,r]` are
    linked to the beads `k2[l,r]`, respectively. This network structure can
    represent arbitrary "homologously linked" polymers, i.e. those of the form

    .. code-block:: text

        --  ---  -------  --  -
          \/   \/       \/  \/
          /\   /\       /\  /\
        --  ---  -------  --  -

    Notice in the example below that `loop_list[1:,1]` and `loop_list[:-1,0]`
    are both the same as `homolog_list`.

    Example
    -------
    Here's an example for a typical case. With 101 beads per polymer (so 100
    inter-bead segments), and one tenth of the beads bound to each other, we
    might get the following Poisson-space homolog points:

    >>> N = 101; FP = 0.1

    >>> homolog_points = np.where(np.random.rand(int(N)) < FP)[0]

    >>> homolog_points
    array([ 5, 18, 27, 59, 68, 82, 90, 94])

    These eight homologous junctions would lead to a simulation array length of
    `2*N - len(homolog_points) = 202 - 8 = 194`. So `N_tot` will be 194, and
    the `loop_list` will look like

    >>> N_tot, loop_list = rouse.homolog_points_to_loops_list(N, homolog_points)

    >>> loop_list
    array([[ -1,   5, 101, 105],
            [  5,  18, 106, 117],
            [ 18,  27, 118, 125],
            [ 27,  59, 126, 156],
            [ 59,  68, 157, 164],
            [ 68,  82, 165, 177],
            [ 82,  90, 178, 184],
            [ 90,  94, 185, 187],
            [ 94, 101, 188, 193]])

    This can be read as follows. The first four beads of each polymer are
    "free", so first row is saying that beads 0 thru 4 correspond to unlinked
    beads in polymer 1. bead 5 is the linked bead (has half the diffusivity).
    since the polymer is of length 101, if there were no connections, the first
    polymer would extend from bead 0 to bead 100, and the second from bead 101
    to bead 201. but since bead "5" is linked to (what would be) bead 106, that
    bead is omitted from the second polymer, and beads 101-105 are the "free
    end" of the second polymer.

    The second row corresponds to the first actual "loop". The two tethering
    points are beads 5 and 18 of the first polymer. What would have been beads
    106 and 119 (that's beads 5 and 18 of the second polymer) are omitted from
    the list of simulated beads. So bead 106 is linked to bead 5 and bead 117
    is linked to bead 18, and beads 106-117 are linked sequentially.

    Similarly, beads 18 and 118 (and 27 and 125) are linked, with beads 118-125
    being linked sequentially.
    """
    homolog_points = np.sort(np.array(homolog_points))
    num_points = len(homolog_points)
    num_loops = num_points + 1 # includes two end non-"loops" (aka free ends)
    loop_list = np.zeros((num_loops, 4))
    if num_loops == 1:
        return 2*int(N), np.array([[-1,N,N,2*N-1]]).astype(int)
    # keep track of length of "x0"
    curr_n = N
    # there's at least one connection point, thanks to num_loops==1 check above
    for i in range(0,num_points+1):
        k1l = homolog_points[i-1] if i > 0 else -1
        k1r = homolog_points[i] if i < num_points else N
        k2l = curr_n
        beads_in_loop = k1r - k1l - 1
        curr_n += beads_in_loop
        k2r = curr_n - 1
        loop_list[i,:] = np.array([k1l, k1r, k2l, k2r])
    return curr_n, loop_list.astype(int)

def split_homologs_X(X, N, loop_list):
    """Make the (Nt,Ntot,3) output of rouse_homologs to (2,Nt,N,3)."""
    N_t, N_tot, _ = X.shape
    X1 = X[:,:N,:].copy()
    X2 = np.zeros_like(X1)
    num_loops, _ = loop_list.shape
    if num_loops == 1:
        X2 = X[:,N:,:]
        return X1, X2
    curr_i2 = 0
    curr_i0 = N
    for i in range(num_loops):
        k1l, k1r, k2l, k2r = loop_list[i]
        num_beads = k1r - k1l - 1
        X2[:,curr_i2:curr_i2+num_beads] = X[:,curr_i0:curr_i0+num_beads]
        curr_i2 += num_beads
        if i < num_loops - 1:
            X2[:,curr_i2] = X[:,k1r]
            curr_i2 += 1
        curr_i0 += num_beads
    return X1, X2

def split_homolog_x(x0, N, loop_list):
    """Make an (N_tot,3) array from rouse_homologs into (2,N,3)."""
    N_tot, _ = x0.shape
    x1 = x0[:N,:].copy()
    x2 = np.zeros((N,3))
    num_loops, _ = loop_list.shape
    if num_loops == 1:
        x2 = x0[N:,:]
        return x1, x2
    curr_i2 = 0
    curr_i0 = N
    for i in range(num_loops):
        k1l, k1r, k2l, k2r = loop_list[i]
        num_beads = k1r - k1l - 1
        x2[curr_i2:curr_i2+num_beads] = x0[curr_i0:curr_i0+num_beads]
        curr_i2 += num_beads
        if i < num_loops - 1:
            x2[curr_i2] = x0[k1r]
            curr_i2 += 1
        curr_i0 += num_beads
    return x1, x2


def sim_loc(i, N, loop_list):
    """
    Return actual indices for bead "i" of the second polymer.

    Useful for indexing directly into the trucated simulation output array.

    Parameters
    ----------
    i : (M,) int, array_like
        beads index is requested for
    N : int
        actual number of beads in each polymer
    loop_list : (~N*FP, 4) int
        connectivity of the homologous polymers being simulated

    Returns
    -------
    iloc : (M,) Optional<int>, array_like
        index in simulation array (or None if the bead is a paired bead)
        for each bead in `i`.
    """
    iloc = np.array(i).copy()
    homolog_points = loop_list[1:,0]
    is_homolog = np.isin(iloc, homolog_points)
    if np.any(is_homolog):
        iloc = iloc.astype(object)
        iloc[is_homolog] = None
    for ii in np.where(~is_homolog)[0]:
        iloc[ii] = N + i[ii] - np.sum(homolog_points < i[ii])
    return iloc


def rouse(N, FP, L, b, D, Aex, rx, ry, rz, t, t_save=None, tether_list=None):
    r"""
    BD simulation of two homologous yeast chromosomes in meiosis.

    An arbitrary elliptical-shaped confinement can be chosen, and in order to
    simulate synaptonemal formation in prophase, some fraction `FP` of the
    "homologous" beads of the polymer can be "rigidly tethered" to each other.
    The simulation treats these pairs as being one bead (with half the
    diffusion coefficient). The elastic force can then be communicated between
    the two polymers via the connecting loci.

    Depending on what part of prophase is being simulated, either (or both) of
    the telomeres or centromeres can be tethered to the confinement by
    including them in `tether_list`. Currently the confinement force and the
    tethering force are both cubic w.r.t. the distance from the confining
    ellipse with the same strength (`Aex`), and have the form described in
    :func:`f_conf_ellipse` (and :func:`f_tether_ellipse`).


    Parameters
    ----------
    N : int
        Number of beads to use in the discretized Rouse chain
    FP : float
        `FP`:math:`\in[0,1] is fraction of beads in the chain (out of `N`) that
        will be tethered to each other
    L : float
        Length of the chain (in desired length units)
    b : float
        Kuhn length of the chain (in same length units as L)
    D : float
        The diffusivity of a Kuhn length of the chain. This can often difficult
        to measure directly, but see the documentation for the function
        :func:`measured_D_to_rouse` for instructions on how to compute this
        value given clean measurements of short-time regular diffusion or
        Rouse-like MSD behavior
    Aex : float
        multiplicative prefactor controlling strength of confinment and
        tethering forces
    rx, ry, rz : float
        radius of confinement in x, y, z direction(s) respectively
    t : (M,) float, array_like
        The BD integrator will step time from t[0] to t[1] to t[2] to ...
    t_save : (m,) float, array_like
        `t_save`:math:`\subset``t` will be the time steps where the simulation
        output is saved. Default is `t_save = t`.
    tether_list : int, array_like
        list of bead indices that will be tethered to the confinement surface,
        for now, both polymers must be tethered at the same beads (but this
        would not be hard to change).

    Returns
    -------
    tether_list : List<int>
        Indices (into compressed `Ntot` output) that were tethered to the
        confinement surface
    loop_list : (~N*FP, 4) int
        Output of `points_to_loops_list` used to specify which points are
        tethered
    X : (len(t_save), Ntot, 3)
        output of BD simulation at each time in t_save. The number of output
        beads Ntot <= 2*N excludes the beads on the second polymer whose
        positions are determined by the fact that they are tethered to the
        first polymer. Use :func:`split_homologs_X` to reshape this output into
        shape `(2, len(t_save), N, 3)`.
    """
    if t_save is None:
        t_save = t

    homolog_points = np.where(np.random.rand(int(N)) < FP)[0]
    N_tot, loop_list = points_to_loops_list(N, homolog_points)

    # homolog points aren't in the simulation array for the second polymer
    tethered_p2 = np.setdiff1d(tether_list, loop_list[1:,0])
    # calculate their actual positions in the simulation array
    tether_list = list(tether_list) + list(sim_loc(tethered_p2, N, loop_list))
    tether_list = np.sort(np.array(tether_list)).astype(int)

    x = _jit_rouse_homologs(int(N), int(N_tot), tether_list, loop_list, L, b, D, Aex, rx, ry, rz, t, t_save)
    return tether_list, loop_list, x


@jit(nopython=True)
def _jit_rouse_homologs(N, N_tot, tether_list, loop_list, L, b, D, Aex, rx, ry,
                        rz, t, t_save, x0=None):
    """
    "Inner loop" of rouse_homologs.

    Some typing stuff: needs N, N_tot as int, needs tether_list/loop_list as
    dtype == int64, and uses loop_list.shape[0] to get num_loops. This is in
    particular important for empty list cases, where you need to do stuff like

    >>> x = rouse._jit_rouse_homologs(int(N), int(2*N),
    >>>         np.array([]).astype(int),
    >>>         np.array([[]]).T.astype(int),
    >>>         L, b, D, Aex, R, R, R, t, t_save)

    """
    N = int(N)
    N_tot = int(N_tot)
    rtol = 1e-5
    # derived parameters
    L0 = L/(N-1) # length per bead
    bhat = np.sqrt(L0*b) # mean squared bond length of discrete gaussian chain
    Nhat = L/b # number of Kuhn lengths in chain
    Dhat = D*N/Nhat # diffusion coef of a discrete gaussian chain bead
    k_over_xi = 3*Dhat/bhat**2
    # initial position
    if x0 is None:
        x0 = init_homolog_rouse_conf(N, N_tot, loop_list, bhat, rx, ry, rz)
    # pre-alloc output
    if t_save is None:
        t_save = t
    x = np.zeros(t_save.shape + x0.shape)
    # setup for saving only requested time points
    save_i = 0
    if 0 == t_save[save_i]:
        x[0] = x0
        save_i += 1
    # extract from loop_list for repeated use in loop
    homolog_points = loop_list[1:, 0]
    # at each step i, we use data (x,t)[i-1] to create (x,t)[i]
    # in order to make it easy to pull into a new functin later, we'll call
    # t[i-1] "t0", old x (x[i-1]) "x0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = t[i] - t[i-1]
        dW = np.random.randn(*x0.shape)
        # -1 or 1, p=1/2
        S = 2*(np.random.rand() < 0.5) - 1
        # D = sigma^2/2 ==> sigma = np.sqrt(2*D)
        Fbrown = np.sqrt(2*Dhat/h)*(dW - S)
        # "homolog paired" beads have half the diffusivity
        Fbrown[homolog_points] = Fbrown[homolog_points]/2
        # estimate for slope at interval start
        K1 = f_conf_ellipse(x0, Aex, rx, ry, rz) + f_elas_homolog_rouse(x0, N, loop_list, k_over_xi) + Fbrown
        K1[tether_list] += f_tether_ellipse(x0[tether_list], Aex, rx, ry, rz)
        Fbrown = np.sqrt(2*Dhat/h)*(dW + S)
        # "homolog paired" beads have half the diffusivity
        Fbrown[homolog_points] = Fbrown[homolog_points]/2
        x1 = x0 + h*K1
        # estimate for slope at interval end
        K2 = f_conf_ellipse(x1, Aex, rx, ry, rz) + f_elas_homolog_rouse(x1, N, loop_list, k_over_xi) + Fbrown
        K2[tether_list] += f_tether_ellipse(x1[tether_list], Aex, rx, ry, rz)
        # average the slope estimates
        x0 = x0 + h * (K1 + K2)/2
        if np.abs(t[i] - t_save[save_i]) < rtol*np.abs(t_save[save_i]):
            x[save_i] = x0
            save_i += 1
    return x


def plot_homologs(X1, X2, loop_list, ax=None):
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
    points = ax.scatter(*X1[loop_list[:-1, 1], :].T, c='r', s=50)
    ax.plot(*X1.T)
    ax.plot(*X2.T)
    return points


class HomologViewer(PolymerViewer):

    def __init__(self, r, N, loop_list):
        # will keep track of which *index* in the time axis should be used to
        # draw the current scene
        self.t = 0
        self.N = N
        self.loop_list = loop_list
        self.points = None
        # define the actual axes layout
        self.fig = plt.figure()
        # area to plot polymers
        self.ax3d = plt.axes([0.05, 0.15, 0.9, 0.8], facecolor='w', projection='3d')
        self.r = r
        # area to plot "slider" for selecting what time to plot
        self.ax = plt.axes([0.1, 0.025, 0.8, 0.05], facecolor='lightgoldenrodyellow')
        self.update_ax_limits()
        # slider to control what "time" is plotted, (0,1) is rescaled total
        # simulation time units
        self.slider_t = Slider(self.ax, 't', 0, 1, valinit=0)
        # handler to check if the "time" slider has been moved
        self.slider_t.on_changed(self.update_t)
        plt.show()

    @property
    def r(self):
        return self._r

    @r.setter
    def r(self, new_r):
        self._r = new_r
        self._x1, self._x2 = split_homologs_X(self._r, self.N, self.loop_list)
        self._num_time_points, self._num_beads, self._d = self._r.shape
        if hasattr(self, 'ax3d'):
            self.update_ax_limits()
            self.update_drawing()

    # should never be called before a Sim has been loaded successfully
    def update_drawing(self):
        num_polymers = 2
        # if num_polymers != len(self.ax3d.lines):
        self.ax3d.lines = []
        self.ax3d.plot(*self._x1[self.t].T, c='b')
        self.ax3d.plot(*self._x2[self.t].T, c='g')
        if self.points is not None:
            self.points.remove()
            del self.points
        self.points = self.ax3d.scatter(*self._x1[self.t,self.loop_list[:-1,1],:].T, c='r', s=50)
        # for i, r in enumerate([self._x1, self._x2]):
        #     x, y, z = (self.r[self.t,:,0], self.r[self.t,:,1],
        #                self.r[self.t,:,2])
        #     self.ax3d.lines[i].set_data(x, y)
        #     self.ax3d.lines[i].set_3d_properties(z)
        #     self.points.remove()
        #     del self.points
        #     self.points = self.ax3d.scatter(*self._x1[self.t,self.loop_list[:-1,1],:].T, c='r', s=50)
        self.fig.canvas.draw_idle()

    def update_ax_limits(self):
        self.ax3d.set_xlim((np.nanmin(self.r[:,:,0]), np.nanmax(self.r[:,:,0])))
        self.ax3d.set_ylim((np.nanmin(self.r[:,:,1]), np.nanmax(self.r[:,:,1])))
        self.ax3d.set_zlim((np.nanmin(self.r[:,:,2]), np.nanmax(self.r[:,:,2])))

    def update_t(self, val):
        if self.r is None:
            return
        # calculate value on slider
        new_t = int(round(val*self._num_time_points))
        # bound it to 0, num_time_points-1
        new_t = max(0, min(self._num_time_points-1, new_t))
        self.t = new_t
        self.update_drawing()






import numpy as np
from scipy.integrate import simps
from matplotlib import pyplot as plt

def invLaplace(p_values, G_values, L, reduction=0.0):
    r"""
    Numerical Inverse Laplace from :math:`G(p) \to G(L)`.

    Parameters
    ----------
    p_values : array_like
        1D complex array of p values (i.e. the contour)
    G_values : array_like
        1D complex array G(p)
    L : float
        Polymer length (In Kuhn lengths)
    reduction : float
        multiply answer by exp(-reduction) for better numerics
    """
    x_values = np.imag(p_values)
    # print(x_values[0:10])
    y_values = np.array(G_values)*np.exp(p_values*L-reduction)
    return simps(y_values, x_values)/(2*np.pi)

def concentrate_high(low, high, halfpoints):
    if low>high:
        return np.flip(concentrate_high(high, low, halfpoints),0)
    mid = 0.9*high+0.1*low
    upps = np.linspace(mid, high, halfpoints)
    lows = np.linspace(low, mid, halfpoints)
    return np.concatenate((lows, upps[1:]))


def make_path(factor=100, scale=1.0, lambda_offset=0.1, width=2.0, depth=2.0,
              nwing=100, nside=30, nback=20, maxp=6000, pole_offset=0.0):
    """
    Path for complex integral that draws a half rectangle about the origin.

    Parameters
    ----------
    factor : int
        Number of points to caltulate (sort of)
    scale : float
        Scale of path.  Recomended: around 10.0/N
    lambda_offset : float
        distance from complex axis
    width : float
        width of rectangle
    depth : float
        depth of rectangle
    nwing : int
        number of points in wing before ``factor``
    nside : int
        number of points to a side before ``factor``
    back : int
        number of points on the back before ``factor``
    maxp : float
        approximation for infinity
    pole_offset : float
        add to depth
    """
    assert depth >= lambda_offset
    nside=nside*factor
    nback=nback*factor
    nwing=nwing*factor
    depth=depth*scale+pole_offset
    lambda_offset=lambda_offset*scale
    maxp=maxp*scale
    width=width*scale


    lower = np.linspace(-maxp, -width, nwing)*1j+lambda_offset

    low_side = -width*1j+np.linspace(lambda_offset,depth,nside)
    #low_side = -width*1j+concentrate_high(lambda_offset,depth,int(nside/2))

    back = np.linspace(-width, width, nback)*1j+depth

    top_side = width*1j+np.linspace(depth, lambda_offset, nside)
    #top_side = width*1j+concentrate_high(depth, lambda_offset, int(nside/2))

    upper = np.linspace(width, maxp, nwing)*1j+lambda_offset

    p_values = np.concatenate((lower, low_side, back, top_side, upper))

    cutts = {'lower':[0, nwing],
             'low_side':[nwing, nwing+nside],
             'back':[nwing+nside, nwing+nside+nback],
             'top_side':[nwing+nside+nback, nwing+2*nside+nback],
             'upper':[nwing+2*nside+nback, 2*nwing+2*nside+nback]}

    return [p_values, cutts]

def plot_path(path):
    plt.plot(np.real(path), np.imag(path),'.-')
    plt.axis('equal')
    plt.show()

def invLaplace_path(path, cutts, G_values, L, reduction=0.0):
    '''Inverse laplace transform based on path.

    Args:
        path (ndarray): complex path values
        cutts (dict): specified by path
        G_values
    '''
    total=0.0+0.0j
    for key, cutt in cutts.items():
        p_values=path[cutt[0]:cutt[1]]
        y_values=G_values[cutt[0]:cutt[1]]*np.exp(p_values*L-reduction)

        if key == 'low_side':
            x_values=np.real(p_values)
            sgn=1.0/(1j*2*np.pi)
        elif key == 'top_side':
            # Need to reverse to give increasing x_values
            x_values=-np.real(p_values)
            sgn=-1.0/(1j*2*np.pi)
        else:
            x_values=np.imag(p_values)
            sgn=1.0/(2*np.pi) # the 1j/1j cancels

        total = total + sgn*simps(y_values, x_values)

    return total

def plot_int_path(G_values, cutts):
    for key, cutt in cutts.items():
        plt.plot(np.arange(cutt[0],cut[1]), np.real(G_values[cutt[0]:cutt[1]]),
                 label=key)
    plt.show()

def plot_invLaplace_path(path, cutts, G_values, L, title="Inverse Laplace"):
    x_total=0.0
    for key, cutt in cutts.items():
        p_values=path[cutt[0]:cutt[1]]
        y_values=G_values[cutt[0]:cutt[1]]*np.exp(p_values*L)

        if key == 'low_side':
            x_values=np.real(p_values)
            sgn=1.0/(1j*2*np.pi)
        elif key == 'top_side':
            # Need to reverse to give increasing x_values
            x_values=-np.real(p_values)
            sgn=-1.0/(1j*2*np.pi)
        else:
            x_values=np.imag(p_values)
            sgn=1.0/(2*np.pi) # the 1j/1j cancels

        x_values = x_values-x_values[0]+x_total
        x_total=x_values[-1]
        plt.plot(x_values, np.real(y_values), '.-')
        plt.plot(x_values, np.real(G_values[cutt[0]:cutt[1]]), '-')
    plt.title(title)
    plt.show()


def test_inv_laplace():
    p_values=make_path(factor=100)
    nums = p_values*0
    aa = -0.1
    for ii, p in enumerate(p_values):
        nums[ii] = 1.0/(p-aa)

    N_set = {0.05, 0.1, 0.5, 1.0, 10}
    for iN, N in enumerate(N_set):
        print("Correct= %f, Calc=%f"%(np.exp(aa*N),
                                      invLaplace_path(nums, N)))
        plot_invLaplace_path(nums, N, title="N="+str(N))
#test_inv_laplace()

def test_scale():
    N_set = {0.05, 0.1, 0.5, 1.0, 10}
    for iN, N in enumerate(N_set):
        p_values=make_path(factor=100, scale=1.0/N)
        nums = p_values*0
        aa = -0.1
        for ii, p in enumerate(p_values):
            nums[ii] = 1.0/(p-aa)

        print("Correct= %f, Calc=%f"%(np.exp(aa*N),
                                      invLaplace_path(nums, N)))
        plot_invLaplace_path(nums, N, title="N="+str(N))
#test_scale()



"""
Compute results related to (twistable, stretchable, shearable) wormlike chains.

This package can be largely separated into

1. Modules for processing simulation output from our FORTRAN codebase
(:mod:`wlcsim.data`, :mod:`wlcsim.input`)
2. Modules for generating polymer chains at equilibrium (both directly in
:mod:`wlcsim.chains` and using Monte Carlo :mod:`wlcsim.mc`).
3. Modules for simulating diffusing polymers (:mod:`wlcsim.bd`).
4. Modules for plotting these various chains and
their properties (:mod:`wlcsim.plot` and :mod:`wlcsim.plot_wlcsim`).

"""
from . import bd
from . import mc
from . import chains
from . import utils
from . import analytical
from . import special

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
